"""
Routines for plots related to stellar particles

"""

import logging
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from .. import array, filt, units, units as _units
from ..analysis import angmom, profile
from ..sph import Kernel2D, render_spherical_image
from . import sph as plot_sph

logger = logging.getLogger('pynbody.plot.stars')


def _bytscl(arr, mini=0, maxi=10000):
	X = (arr - mini) / (maxi - mini)
	X[X > 1] = 1
	X[X < 0] = 0
	return X


def _nw_scale_rgb(r, g, b, scales=[4, 3.2, 3.4]):
	return r * scales[0], g * scales[1], b * scales[2]


def _nw_arcsinh_fit(r, g, b, nonlinearity=3):
	radius = r + g + b
	val = np.arcsinh(radius * nonlinearity) / nonlinearity / radius
	return r * val, g * val, b * val


def _combine(r, g, b, magnitude_range, brightest_mag=None, mollview=False):
	# flip sign so that brightest pixels have biggest value
	r = -r
	g = -g
	b = -b

	if brightest_mag is None:
		brightest_mag = []

		# find something close to the maximum that is not quite the maximum
		for x in r, g, b:
			if mollview:
				x_tmp = x.flatten()[x.flatten()<0]
				ordered = np.sort(x_tmp.data)
			else:
				ordered = np.sort(x.flatten())
			brightest_mag.append(ordered[-len(ordered) // 5000])

		brightest_mag = max(brightest_mag)
	else:
		brightest_mag = -brightest_mag

	rgbim = np.zeros((r.shape[0], r.shape[1], 3))
	rgbim[:, :, 0] = _bytscl(r, brightest_mag - magnitude_range, brightest_mag)
	rgbim[:, :, 1] = _bytscl(g, brightest_mag - magnitude_range, brightest_mag)
	rgbim[:, :, 2] = _bytscl(b, brightest_mag - magnitude_range, brightest_mag)
	return rgbim, -brightest_mag

def _convert_to_mag_arcsec2(image, mollview=False):
	if not mollview:
		assert image.units=="pc^-2"
	pc2_to_sqarcsec = 2.3504430539466191e-09
	return -2.5*np.log10(image*pc2_to_sqarcsec)

def contour_surface_brightness(sim, band='v', width=50, resolution=None, axes=None, label=True,
								contour_kwargs=None, smooth_min=0.0):
	"""Plot surface brightness contours in the given band.

	For information about how surface brightnesses are calculated, see the documentation for
	:mod:`pynbody.analysis.luminosity`.

	Parameters
	----------

	sim : SimSnap
		The simulation snapshot to plot.

	band : str
		The band to plot. Default is 'v'.

	width : float | str, optional
		The width of the image to produce

	resolution : int, optional
		The resolution of the image to produce. The default is determined by the configuration
		option ``image-default-resolution``.

	axes : matplotlib axes object, optional
		If not None, the axes object to plot to

	label : bool, optional
		If True, the contours are labelled with the surface brightness in mag arcsec^-2

	contour_kwargs : dict or None, optional
		Keyword arguments to pass to the matplotlib contour plot function. For example, this
		can be used to determine the contour levels.

	smooth_min : float or str, optional
		The minimum size of the smoothing kernel, either as a float or a unit string.
		Setting this to a non-zero value makes smoother, clearer contours but loses fine detail.
		Default is 0.0.

	"""

	return plot_sph.contour(sim, qty=band + '_lum_den', units="pc^-2", width=width,
							resolution=resolution, axes=axes, label=label, smooth_min=smooth_min,
							_transform = _convert_to_mag_arcsec2,
							log=False, contour_kwargs=contour_kwargs)



def render(sim,
		   r_band='i', g_band='v', b_band='u',
		   width=50,
		   r_scale=0.5, g_scale=1.0, b_scale=1.0,
		   with_dust=False,
		   dynamic_range=2.0,
		   mag_range=None,
		   resolution=None,
		   noplot=False, axes=None, return_image=False, clear=True,
		   return_range=False):
	'''
	Make a 3-color image of stars.

	For information about how surface brightnesses are calculated, see the documentation for
	:mod:`pynbody.analysis.luminosity`.

	If ``with_dust`` is True, a simple dust screening is applied; see below for important notes
	on the dust screening.

	Parameters
	----------

	sim : SimSnap
		The simulation snapshot to plot.

	r_band, g_band, b_band : str
		The filter bands to use for R, G and B image channels. Default is 'i', 'v', 'b'.

	width : float | str, optional
		The width of the image to produce, either as a float or a unit string.

	r_scale, g_scale, b_scale : float, optional
		The scaling of the red, green and blue channels before they are combined.

	dynamic_range : float, optional
		The number of dex in luminosity over which the image brightness ranges, if
		``mag_range`` is not provided.

	with_dust : bool, optional
		If True, the image is rendered with a simple dust screening model. See important notes below.

	mag_range : tuple, optional
		The brightest and faintest surface brightnesses in the final image, in
		mag arcsec^-2.

	resolution : int, optional
		The resolution of the image to produce. The default is determined by the configuration
		option ``image-default-resolution``.

	noplot : bool, optional
		If True, the image is not plotted; most useful alongside ``return_image``, if you want to save the
		image to a file.

	axes : matplotlib.axes.Axes, optional
		If not None, the axes object to plot to.

	return_image : bool, optional
		If True, the image is returned as an array (N x N x 3) for the RGB channels. Default is False.

	return_range : bool, optional
		If True, a tuple with the range of the image in mag arcsec^-2 is returned. Default is False.

	clear : bool, optional
		If True, the current figure is cleared before plotting. Default is True.


	Returns
	-------

	If ``return_image`` is True, an array (N x N x 3) representing the RGB image is returned.

	If ``return_range`` is True, a tuple with the range of the image in mag arcsec^-2 is returned.

	Notes
	-----

	The dust screening model is exceptionally simple and can only be used for indicative purposes.
	For more accurate results, radiative transfer is essential and is provided by other packages
	such as `skirt <https://skirt.ugent.be/root/_home.html>`_.

	The model assumes that the dust is proportional to the metal density. It estimates a V-band
	extinction A_V using empirical data from Draine & Lee (1984, ApJ, 285, 89) and Savage and Mathis
	(1979, ARA&A, 17, 73). This is then converted to extinction in the given bands using the
	Calzetti law.

	The model furthermore assumes that half the dust is in front of the stars and half behind, because
	there is no radiative transfer to account for the actual distribution of dust in the 3d space.

	'''

	width = plot_sph._width_in_sim_units(sim, width)

	r = plot_sph.image(sim.s, qty=r_band + '_lum_den', width=width, log=False,
			  units="pc^-2", clear=False, noplot=True, resolution=resolution) * r_scale
	g = plot_sph.image(sim.s, qty=g_band + '_lum_den', width=width, log=False,
			  units="pc^-2", clear=False, noplot=True, resolution=resolution) * g_scale
	b = plot_sph.image(sim.s, qty=b_band + '_lum_den', width=width, log=False,
			  units="pc^-2", clear=False, noplot=True, resolution=resolution) * b_scale

	# convert all channels to mag arcsec^-2

	r=_convert_to_mag_arcsec2(r)
	g=_convert_to_mag_arcsec2(g)
	b=_convert_to_mag_arcsec2(b)

	if with_dust is True:
		# render image with a simple dust absorption correction based on Calzetti's law using the gas content.

		a_v = _dust_Av_image(sim, width, resolution)

		try:
			import extinction
		except ImportError:
			raise ImportError("Could not load extinction package. If you want to use this feature, "
				"plaese install the extinction package from here: http://extinction.readthedocs.io/en/latest/"
				"or <via pip install extinction> or <conda install -c conda-forge extinction>") from None

		#get the central wavelength for the given band
		wavelength_avail = {'u':3571,'b':4378,'v':5466,'r':6695,'i':8565,'j':12101,
				   'h':16300,'k':21900,'U':3571,'B':4378,'V':5466,'R':6695,'I':8565,'J':12101,
				   'H':16300,'K':21900} #in Angstrom
		# effective wavelength taken from http://svo2.cab.inta-csic.es/svo/theory/fps3/index.php?mode=browse&gname=Generic&gname2=Johnson
		# and from https://en.wikipedia.org/wiki/Photometric_system for h, k

		lr,lg,lb = wavelength_avail[r_band],wavelength_avail[g_band],wavelength_avail[b_band] #in Angstrom
		wave = np.array([lb, lg, lr])

		ext_r = np.empty(a_v.shape)
		ext_g = np.empty(a_v.shape)
		ext_b = np.empty(a_v.shape)

		for i in range(len(a_v)):
			for j in range(len(a_v[0])):
				ext = extinction.calzetti00(wave.astype(np.float_), a_v[i][j].astype(np.float_), 3.1, unit='aa', out=None)
				ext_r[i][j] = ext[2]
				ext_g[i][j] = ext[1]
				ext_b[i][j] = ext[0]

		r = r+ext_r
		g = g+ext_g
		b = b+ext_b

	#r,g,b = nw_scale_rgb(r,g,b)
	#r,g,b = nw_arcsinh_fit(r,g,b)

	if mag_range is None:
		rgbim, mag_max = _combine(r, g, b, dynamic_range * 2.5)
		mag_min = mag_max + 2.5*dynamic_range
	else:
		mag_max, mag_min = mag_range
		rgbim, mag_max = _combine(r, g, b, mag_min - mag_max, mag_max)

	if not noplot:
		if clear:
			plt.clf()
		if axes is None:
			axes = plt.gca()

		if axes:
			axes.imshow(
				rgbim[::-1, :], extent=(-width / 2, width / 2, -width / 2, width / 2))
			axes.set_xlabel('x [' + str(sim.s['x'].units) + ']')
			axes.set_ylabel('y [' + str(sim.s['y'].units) + ']')
			plt.draw()

	if return_image:
		return rgbim

	if return_range:
		return mag_max, mag_min



def _dust_Av_image(sim, width, resolution):
	"""Produce a map of the extinction Av for the given simulation, using the gas content.

	Note that the dust model is very simple and naive! See comments inline.
	"""

	# Assume that only the gas with z>0 absorbs light (i.e. 'all light' is produced in the midplane)
	gas = sim.g

	# calculate the density of metals, and assume that dust is proportional to the metal density
	rho_metals = gas['rho'] * gas['metals']
	rho_metals.units = gas['rho'].units
	column_den = plot_sph.image(gas, qty=rho_metals, width=width, log=False, restrict_depth=True,
					            units="m_p cm^-2", clear=False, noplot=True, resolution=resolution)

	# From Draine & Lee (1984, ApJ, 285, 89) in the V band (lambda^-1 ~= 2 micron^-1), the optical
	# depth is 0.5 for an H column density of 10^21 cm^2. That scaling in turn is based on data in
	# the review of Savage and Mathis (1979, ARA&A, 17, 73). Amazingly, this is the most up-to-date
	# reference I could find on the subject. There is no mention of the metallicity dependence but
	# one would assume dust columns should scale at least with metal columns (perhaps even more
	# steeply with local metallicity). Given the wild approximations in all this, I assume a
	# metallicity of 0.02 for gas in the Milky Way

	tau_to_mag_extinction = 2.5 / np.log(10.)

	a_v = tau_to_mag_extinction * 0.5 * column_den / 1e21 / 0.02

	# Finally, we assume that half the dust is in front of the stars and half behind. This is
	# yet another big assumption!

	a_v/=2

	return a_v


def mollview(map=None,fig=None,plot=False,filenme=None,
			 rot=None,coord=None,unit='',
			 xsize=800,title='Mollweide view',nest=False,
			 min=None,max=None,flip='astro',
			 remove_dip=False,remove_mono=False,
			 gal_cut=0,
			 format='%g',format2='%g',
			 cbar=True,cmap=None, notext=False,
			 norm=None,hold=False,margins=None,sub=None,
			 return_projected_map=False):
	"""Plot an healpix map (given as an array) in Mollweide projection.
	   Requires the healpy package.

	   This function is taken from the Healpy package and slightly modified.

	Parameters
	----------
	map : float, array-like or None
	  An array containing the map, supports masked maps, see the `ma` function.
	  If None, will display a blank map, useful for overplotting.
	fig : figure object or None, optional
	  The figure to use. Default: create a new figure
	plot : bool (default: False)
	  if True the image is plotted
	filename : string (default: None)
		 Filename to be written to (if a filename is specified)
	rot : scalar or sequence, optional
	  Describe the rotation to apply.
	  In the form (lon, lat, psi) (unit: degrees) : the point at
	  longitude *lon* and latitude *lat* will be at the center. An additional rotation
	  of angle *psi* around this direction is applied.
	coord : sequence of character, optional
	  Either one of 'G', 'E' or 'C' to describe the coordinate
	  system of the map, or a sequence of 2 of these to rotate
	  the map from the first to the second coordinate system.
	unit : str, optional
	  A text describing the unit of the data. Default: ''
	xsize : int, optional
	  The size of the image. Default: 800
	title : str, optional
	  The title of the plot. Default: 'Mollweide view'
	nest : bool, optional
	  If True, ordering scheme is NESTED. Default: False (RING)
	min : float, optional
	  The minimum range value
	max : float, optional
	  The maximum range value
	flip : {'astro', 'geo'}, optional
	  Defines the convention of projection : 'astro' (default, east towards left, west towards right)
	  or 'geo' (east towards right, west towards left)
	remove_dip : bool, optional
	  If :const:`True`, remove the dipole+monopole
	remove_mono : bool, optional
	  If :const:`True`, remove the monopole
	gal_cut : float, scalar, optional
	  Symmetric galactic cut for the dipole/monopole fit.
	  Removes points in latitude range [-gal_cut, +gal_cut]
	format : str, optional
	  The format of the scale label. Default: '%g'
	format2 : str, optional
	  Format of the pixel value under mouse. Default: '%g'
	cbar : bool, optional
	  Display the colorbar. Default: True
	notext : bool, optional
	  If True, no text is printed around the map
	norm : {'hist', 'log', None}
	  Color normalization, hist= histogram equalized color mapping,
	  log= logarithmic color mapping, default: None (linear color mapping)
	hold : bool, optional
	  If True, replace the current Axes by a MollweideAxes.
	  use this if you want to have multiple maps on the same
	  figure. Default: False
	sub : int, scalar or sequence, optional
	  Use only a zone of the current figure (same syntax as subplot).
	  Default: None
	margins : None or sequence, optional
	  Either None, or a sequence (left,bottom,right,top)
	  giving the margins on left,bottom,right and top
	  of the axes. Values are relative to figure (0-1).
	  Default: None
	return_projected_map : bool
	  if True returns the projected map in a 2d numpy array
	See Also
	--------
	gnomview, cartview, orthview, azeqview
	"""
	try:
		from healpy import pixelfunc, projaxes as PA
	except ImportError:
		warnings.warn(
			"Could not load healpy package. If you want to use this feature, "
			"plaese install the healpy package from here: http://healpy.readthedocs.io/en/latest/"
			"or via pip or conda.", RuntimeWarning)
		return

	# Create the figure

	if not (hold or sub):
		if fig is None:
			f=plt.figure(figsize=(8.5,5.4))
			extent = (0.02,0.05,0.96,0.9)
		else:
			f=fig
			extent = (0.02,0.05,0.96,0.9)
	elif hold:
		f=plt.gcf()
		left,bottom,right,top = np.array(f.gca().get_position()).ravel()
		extent = (left,bottom,right-left,top-bottom)
		f.delaxes(f.gca())
	else: # using subplot syntax
		f=plt.gcf()
		if hasattr(sub,'__len__'):
			nrows, ncols, idx = sub
		else:
			nrows, ncols, idx = sub//100, (sub%100)//10, (sub%10)
		if idx < 1 or idx > ncols*nrows:
			raise ValueError('Wrong values for sub: %d, %d, %d'%(nrows,
															 ncols,
															 idx))
		c,r = (idx-1)%ncols,(idx-1)//ncols
		if not margins:
			margins = (0.01,0.0,0.0,0.02)
		extent = (c*1./ncols+margins[0],
			  1.-(r+1)*1./nrows+margins[1],
			  1./ncols-margins[2]-margins[0],
			  1./nrows-margins[3]-margins[1])
		extent = (extent[0]+margins[0],
			  extent[1]+margins[1],
			  extent[2]-margins[2]-margins[0],
			  extent[3]-margins[3]-margins[1])

	# Starting to draw : turn interactive off
	wasinteractive = plt.isinteractive()
	plt.ioff()
	try:
		if map is None:
			map = np.zeros(12)+np.inf
			cbar=False
		map = pixelfunc.ma_to_array(map)
		ax=PA.HpxMollweideAxes(f,extent,coord=coord,rot=rot,
						   format=format2,flipconv=flip)
		f.add_axes(ax)
		if remove_dip:
			map=pixelfunc.remove_dipole(map,gal_cut=gal_cut,
									nest=nest,copy=True,
									verbose=True)
		elif remove_mono:
			map=pixelfunc.remove_monopole(map,gal_cut=gal_cut,nest=nest,
									  copy=True,verbose=True)
		img = ax.projmap(map,nest=nest,xsize=xsize,coord=coord,vmin=min,vmax=max,
			   cmap=cmap,norm=norm)
		if cbar:
			im = ax.get_images()[0]
			b = im.norm.inverse(np.linspace(0,1,im.cmap.N+1))
			v = np.linspace(im.norm.vmin,im.norm.vmax,im.cmap.N)
			if matplotlib.__version__ >= '0.91.0':
				cb=f.colorbar(im,ax=ax,
						  orientation='horizontal',
						  shrink=0.5,aspect=25,ticks=PA.BoundaryLocator(),
						  pad=0.05,fraction=0.1,boundaries=b,values=v,
						  format=format)
			else:
				# for older matplotlib versions, no ax kwarg
				cb=f.colorbar(im,orientation='horizontal',
						  shrink=0.5,aspect=25,ticks=PA.BoundaryLocator(),
						  pad=0.05,fraction=0.1,boundaries=b,values=v,
						  format=format)
			cb.solids.set_rasterized(True)
		ax.set_title(title)
		if not notext:
			ax.text(0.86,0.05,ax.proj.coordsysstr,fontsize=14,
				fontweight='bold',transform=ax.transAxes)
		if cbar:
			cb.ax.text(0.5,-1.0,unit,fontsize=14,
				   transform=cb.ax.transAxes,ha='center',va='center')
		f.sca(ax)
	finally:
		if plot:
			plt.draw()
		if wasinteractive:
			plt.ion()
			#plt.show()
	if return_projected_map:
		return img



def render_mollweide(sim, filename=None,
		   r_band='i', g_band='v', b_band='u',
		   r_scale=0.5, g_scale=1.0, b_scale=1.0,
		   dynamic_range=2.0,
		   mag_range=None,
		   width=25,
		   nside=128,
		   starsize=None,
		   plot=True, axes=None, ret_im=False, clear=True,
		   ret_range=False):
	'''
	Make a 3-color all-sky image of stars in a mollweide projection.
	Adapted from the function pynbody.plot.stars.render

	The colors are based on magnitudes found using stellar Marigo
	stellar population code.  However there is no radiative transfer
	to account for dust.

	Returns: If ret_im=True, an NxNx3 array representing an RGB image

	**Optional keyword arguments:**

	   *filename*: string (default: None)
		 Filename to be written to (if a filename is specified)

	   *r_band*: string (default: 'i')
		 Determines which Johnston filter will go into the image red channel

	   *g_band*: string (default: 'v')
		 Determines which Johnston filter will go into the image green channel

	   *b_band*: string (default: 'b')
		 Determines which Johnston filter will go into the image blue channel

	   *r_scale*: float (default: 0.5)
		 The scaling of the red channel before channels are combined

	   *g_scale*: float (default: 1.0)
		 The scaling of the green channel before channels are combined

	   *b_scale*: float (default: 1.0)
		 The scaling of the blue channel before channels are combined

	   *width*: float in kpc (default:50)
		 Sets the size of the image field in kpc

	   *starsize*: float in kpc (default: None)
		 If not None, sets the maximum size of stars in the image

	   *ret_im*: bool (default: False)
		 if True, the NxNx3 image array is returned

	   *ret_range*: bool (default: False)
		 if True, the range of the image in mag arcsec^-2 is returned.

	   *plot*: bool (default: True)
		 if True, the image is plotted

	   *axes*: matplotlib axes object (deault: None)
		 if not None, the axes object to plot to

	   *dynamic_range*: float (default: 2.0)
		 The number of dex in luminosity over which the image brightness ranges

	   *mag_range*: float, float (default: None)
		 If provided, the brightest and faintest surface brightnesses in the range,
		 in mag arcsec^-2. Takes precedence over dynamic_range.
	'''

	if isinstance(width, str) or issubclass(width.__class__, _units.UnitBase):
		if isinstance(width, str):
			width = _units.Unit(width)
		width = width.in_units(sim['pos'].units, **sim.conversion_context())

	if starsize is not None:
		smf = filt.HighPass('smooth', str(starsize) + ' kpc')
		sim.s[smf]['smooth'] = array.SimArray(starsize, 'kpc', sim=sim)


	r = render_spherical_image(sim.s, qty=r_band + '_lum_den', nside=nside, distance=width, kernel=Kernel2D(),kstep=0.5, denoise=None, out_units="pc^-2", threaded=False)# * r_scale
	r = mollview(r,return_projected_map=True) * r_scale
	f=plt.gcf()
	g = render_spherical_image(sim.s, qty=g_band + '_lum_den', nside=nside, distance=width, kernel=Kernel2D(),kstep=0.5, denoise=None, out_units="pc^-2", threaded=False)# * g_scale
	g = mollview(g,return_projected_map=True,fig=f) * g_scale
	f=plt.gcf()
	b = render_spherical_image(sim.s, qty=b_band + '_lum_den', nside=nside, distance=width, kernel=Kernel2D(),kstep=0.5, denoise=None, out_units="pc^-2", threaded=False)# * b_scale
	b = mollview(b,return_projected_map=True,fig=f) * b_scale
	# convert all channels to mag arcsec^-2

	r=_convert_to_mag_arcsec2(r, mollview=True)
	g=_convert_to_mag_arcsec2(g, mollview=True)
	b=_convert_to_mag_arcsec2(b, mollview=True)

	if mag_range is None:
		rgbim, mag_max = _combine(r, g, b, dynamic_range * 2.5, mollview=True)
		mag_min = mag_max + 2.5*dynamic_range
	else:
		mag_max, mag_min = mag_range
		rgbim, mag_max = _combine(r, g, b, mag_min - mag_max, mag_max, mollview=True)

	if plot:
		if clear:
			plt.clf()
		if axes is None:
			axes = plt.gca()

		if axes:
			axes.imshow(
				rgbim[::-1, :])#, extent=(-width / 2, width / 2, -width / 2, width / 2)
			axes.axis('off')
			plt.draw()

	if filename:
		plt.savefig(filename)

	if ret_im:
		return rgbim

	if ret_range:
		return mag_max, mag_min


def sfh(sim, filename=None, massform=True, clear=False, legend=False,
		subplot=False, trange=False, bins=100, **kwargs):
	'''
	star formation history

	**Optional keyword arguments:**

	   *trange*: list, array, or tuple
		 size(t_range) must be 2. Specifies the time range.

	   *bins*: int
		 number of bins to use for the SFH

	   *massform*: bool
		 decides whether to use original star mass (massform) or final star mass

	   *subplot*: subplot object
		 where to plot SFH

	   *legend*: boolean
		 whether to draw a legend or not

	   *clear*: boolean
		 if False (default), plot on the current axes. Otherwise, clear the figure first.

	By default, sfh will use the formation mass of the star.  In tipsy, this will be
	taken from the starlog file.  Set massform=False if you want the final (observed)
	star formation history

	**Usage:**

	>>> import pynbody.plot as pp
	>>> pp.sfh(s,linestyle='dashed',color='k')


	'''
	import matplotlib.pyplot as pyplot

	if subplot:
		plt = subplot
	else:
		plt = pyplot

	if "nbins" in kwargs:
		bins = kwargs['nbins']

	if 'nbins' in kwargs:
		bins = kwargs['nbins']
		del kwargs['nbins']

	if ((len(sim.g)>0) | (len(sim.d)>0)): simstars = sim.star
	else: simstars = sim

	if trange:
		assert len(trange) == 2
	else:
		trange = [simstars['tform'].in_units(
			"Gyr").min(), simstars['tform'].in_units("Gyr").max()]
	binnorm = 1e-9 * bins / (trange[1] - trange[0])

	trangefilt = filt.And(filt.HighPass('tform', str(trange[0]) + ' Gyr'),
						  filt.LowPass('tform', str(trange[1]) + ' Gyr'))
	tforms = simstars[trangefilt]['tform'].in_units('Gyr')

	if massform:
		try:
			weight = simstars[trangefilt][
				'massform'].in_units('Msol') * binnorm
		except (KeyError, units.UnitsException):
			warnings.warn(
				"Could not load massform array -- falling back to current stellar masses", RuntimeWarning)
			weight = simstars[trangefilt]['mass'].in_units('Msol') * binnorm
	else:
		weight = simstars[trangefilt]['mass'].in_units('Msol') * binnorm

	if clear:
		plt.clf()
	sfhist, thebins, patches = plt.hist(tforms, weights=weight, bins=bins,
										histtype='step', **kwargs)
	if not subplot:
		# don't set the limits
		#plt.ylim(0.0, 1.2 * np.max(sfhist))
		plt.xlabel('Time [Gyr]', fontsize='large')
		plt.ylabel(r'SFR [M$_\odot$ yr$^{-1}$]', fontsize='large')
	else:
		plt.set_ylim(0.0, 1.2 * np.max(sfhist))

	# Make both axes have the same start and end point.
	if subplot:
		x0, x1 = plt.get_xlim()
	else:
		x0, x1 = plt.gca().get_xlim()


	# add a z axis on top if it has not been already done by an earlier plot:
	from pynbody.analysis import pkdgrav_cosmo as cosmo
	c = cosmo.Cosmology(sim=sim)

	old_axis = pyplot.gca()

	pz = plt.twiny()
	labelzs = np.arange(5, int(sim.properties['z']) - 1, -1)
	times = [13.7 * c.Exp2Time(1.0 / (1 + z)) / c.Exp2Time(1) for z in labelzs]
	pz.set_xticks(times)
	pz.set_xticklabels([str(x) for x in labelzs])
	pz.set_xlim(x0, x1)
	pz.set_xlabel('$z$')
	pyplot.sca(old_axis)

	if legend:
		plt.legend(loc=1)
	if filename:
		logger.info("Saving %s", filename)
		plt.savefig(filename)

	return array.SimArray(sfhist, "Msol yr**-1"), array.SimArray(thebins, "Gyr")


def schmidtlaw(sim, center=True, filename=None, pretime='50 Myr',
			   diskheight='3 kpc', rmax='20 kpc', compare=True,
			   radial=True, clear=True, legend=True, bins=10, **kwargs):
	'''Schmidt Law

	Plots star formation surface density vs. gas surface density including
	the observed relationships.  Currently, only plots densities found in
	radial annuli.

	**Usage:**

	>>> import pynbody.plot as pp
	>>> pp.schmidtlaw(h[1])

	**Optional keyword arguments:**

	   *center*: bool
		 center and align the input simulation faceon.

	   *filename*: string
		 Name of output file

	   *pretime* (default='50 Myr'): age of stars to consider for SFR

	   *diskheight* (default='3 kpc'): height of gas and stars above
		  and below disk considered for SF and gas densities.

	   *rmax* (default='20 kpc'): radius of disk considered

	   *compare* (default=True):  whether to include Kennicutt (1998) and
			Bigiel+ (2008) for comparison

	   *radial* (default=True):  should bins be annuli or a rectangular grid?

	   *bins* (default=10):  How many radial bins should there be?

	   *legend*: boolean
		 whether to draw a legend or not
	'''

	if not radial:
		raise NotImplementedError("Sorry, only radial Schmidt law currently supported")

	if center:
		angmom.faceon(sim)

	if isinstance(pretime, str):
		pretime = units.Unit(pretime)

	# select stuff
	diskgas = sim.gas[filt.Disc(rmax, diskheight)]
	diskstars = sim.star[filt.Disc(rmax, diskheight)]

	youngstars = np.where(diskstars['tform'].in_units("Myr") >
						  sim.properties['time'].in_units(
							  "Myr", **sim.conversion_context())
						  - pretime.in_units('Myr'))[0]

	# calculate surface densities
	if radial:
		ps = profile.Profile(diskstars[youngstars], nbins=bins)
		pg = profile.Profile(diskgas, nbins=bins)
	else:
		# make bins 2 kpc
		nbins = rmax * 2 / binsize
		pg, x, y = np.histogram2d(diskgas['x'], diskgas['y'], bins=nbins,
								  weights=diskgas['mass'],
								  range=[(-rmax, rmax), (-rmax, rmax)])
		ps, x, y = np.histogram2d(diskstars[youngstars]['x'],
								  diskstars[youngstars]['y'],
								  weights=diskstars['mass'],
								  bins=nbins, range=[(-rmax, rmax), (-rmax, rmax)])

	if clear:
		plt.clf()

	plt.loglog(pg['density'].in_units('Msol pc^-2'),
			   ps['density'].in_units('Msol kpc^-2') / pretime / 1e6, "+",
			   **kwargs)

	if compare:
		xsigma = np.logspace(np.log10(pg['density'].in_units('Msol pc^-2')).min(),
							 np.log10(
								 pg['density'].in_units('Msol pc^-2')).max(),
							 100)
		ysigma = 2.5e-4 * xsigma ** 1.5        # Kennicutt (1998)
		xbigiel = np.logspace(1, 2, 10)
		ybigiel = 10. ** (-2.1) * xbigiel ** 1.0   # Bigiel et al (2007)
		plt.loglog(xsigma, ysigma, label='Kennicutt (1998)')
		plt.loglog(
			xbigiel, ybigiel, linestyle="dashed", label='Bigiel et al (2007)')

	plt.xlabel(r'$\Sigma_{gas}$ [M$_\odot$ pc$^{-2}$]')
	plt.ylabel(r'$\Sigma_{SFR}$ [M$_\odot$ yr$^{-1}$ kpc$^{-2}$]')
	if legend:
		plt.legend(loc=2)
	if (filename):
		logger.info("Saving %s", filename)
		plt.savefig(filename)


def oneschmidtlawpoint(sim, center=True, pretime='50 Myr',
					   diskheight='3 kpc', rmax='20 kpc', **kwargs):
	'''One Schmidt Law Point

	Determines values for star formation surface density and gas surface
	density for the entire galaxy based on the half mass cold gas radius.

	**Usage:**
	import pynbody.plot as pp
	pp.oneschmidtlawpoint(h[1])

	   *pretime* (default='50 Myr'): age of stars to consider for SFR

	   *diskheight* (default='3 kpc'): height of gas and stars above
		  and below disk considered for SF and gas densities.

	   *rmax* (default='20 kpc'): radius of disk considered
	'''

	if center:
		angmom.faceon(sim)

	cg = h[1].gas[filt.LowPass('temp', 1e5)]
	cgd = cg[filt.Disc('30 kpc', '3 kpc')]
	cgs = np.sort(cgd['rxy'].in_units('kpc'))
	rhgas = cgs[len(cgs) / 2]
	instars = h[1].star[filt.Disc(str(rhgas) + ' kpc', '3 kpc')]
	minstars = np.sum(
		instars[filt.LowPass('age', '100 Myr')]['mass'].in_units('Msol'))
	ingasm = np.sum(
		cg[filt.Disc(str(rhgas) + ' kpc', '3 kpc')]['mass'].in_units('Msol'))
	rpc = rhgas * 1000.0
	rkpc = rhgas
	xsigma = ingasm / (np.pi * rpc * rpc)
	ysigma = minstars / (np.pi * rkpc * rkpc * 1e8)
	return xsigma, ysigma


def satlf(sim, band='v', filename=None, MWcompare=True, Trentham=True,
		  clear=True, legend=True,
		  label='Simulation', **kwargs):
	'''

	satellite luminosity function

	**Options:**

	*band* ('v'): which Johnson band to use. available filters: U, B,
	V, R, I, J, H, K

	*filename* (None): name of file to which to save output

	*MWcompare* (True): whether to plot comparison lines to MW

	*Trentham* (True): whether to plot comparison lines to Trentham +
					 Tully (2009) combined with Koposov et al (2007)

	By default, satlf will use the formation mass of the star.  In
	tipsy, this will be taken from the starlog file.

	**Usage:**

	>>> import pynbody.plot as pp
	>>> h = s.halos()
	>>> pp.satlf(h[1],linestyle='dashed',color='k')


	'''
	import os

	from ..analysis import luminosity as lum

	halomags = []
	# try :
	for haloid in sim.properties['children']:
		if (sim._halo_catalogue.contains(haloid)):
			halo = sim._halo_catalogue[haloid]
			try:
				halo.properties[band + '_mag'] = lum.halo_mag(halo, band=band)
				if np.isfinite(halo.properties[band + '_mag']):
					halomags.append(halo.properties[band + '_mag'])
			except IndexError:
				pass  # no stars in satellite
	# except KeyError:
		#raise KeyError, str(sim)+' properties have no children key as a halo type would'

	if clear:
		plt.clf()
	plt.semilogy(sorted(halomags), np.arange(len(halomags)) + 1, label=label,
				 **kwargs)
	plt.xlabel('M$_{' + band + '}$')
	plt.ylabel('Cumulative LF')
	if MWcompare:
		# compare with observations of MW
		tolfile = os.path.join(os.path.dirname(__file__), "tollerud2008mw")
		if os.path.exists(tolfile):
			tolmags = [float(q) for q in file(tolfile).readlines()]
		else:
			raise OSError(tolfile + " not found")
		plt.semilogy(sorted(tolmags), np.arange(len(tolmags)),
					 label='Milky Way')

	if Trentham:
		halomags = np.array(halomags)
		halomags = halomags[np.asarray(np.where(np.isfinite(halomags)))]
		xmag = np.linspace(halomags.min(), halomags.max(), 100)
		# Trentham + Tully (2009) equation 6
		# number of dwarfs between -11>M_R>-17 is well correlated with mass
		logNd = 0.91 * np.log10(sim.properties['mass']) - 10.2
		# set Nd from each equal to combine Trentham + Tully with Koposov
		coeff = 10.0 ** logNd / (10 ** -0.6 - 10 ** -1.2)

		yn = coeff * 10 ** ((xmag + 5.0) / 10.0)  # Koposov et al (2007)
		plt.semilogy(xmag, yn, linestyle="dashed",
					 label='Trentham & Tully (2009)')

	if legend:
		plt.legend(loc=2)
	if (filename):
		logger.info("Saving %s", filename)
		plt.savefig(filename)


def sbprofile(sim, band='v', diskheight='3 kpc', rmax='20 kpc', binning='equaln',
			  center=True, clear=True, filename=None, axes=False, fit_exp=False,
			  print_ylabel=True, fit_sersic=False, **kwargs):
	'''

	surface brightness profile

	**Usage:**

	>>> import pynbody.plot as pp
	>>> h = s.halos()
	>>> pp.sbprofile(h[1],exp_fit=3,linestyle='dashed',color='k')

	**Options:**

	*band* ('v'): which Johnson band to use. available filters: U, B,
					 V, R, I, J, H, K

	*fit_exp*(False): Fits straight exponential line outside radius specified.

	*fit_sersic*(False): Fits Sersic profile outside radius specified.

	*diskheight('3 kpc')*
	*rmax('20 kpc')*:  Size of disk to be profiled

	*binning('equaln')*:  How show bin sizes be determined? based on
		  pynbody.analysis.profile

	*center(True)*:  Automatically align face on and center?

	*axes(False)*: In which axes (subplot) should it be plotted?

	*filename* (None): name of file to which to save output

	**needs a description of all keywords**

	By default, sbprof will use the formation mass of the star.
	In tipsy, this will be taken from the starlog file.

	'''

	if center:
		logger.info("Centering...")
		angmom.faceon(sim)

	logger.info("Selecting disk stars")
	diskstars = sim.star[filt.Disc(rmax, diskheight)]
	logger.info("Creating profile")
	ps = profile.Profile(diskstars, type=binning)
	logger.info("Plotting")
	r = ps['rbins'].in_units('kpc')

	if axes:
		plt = axes
	else:
		import matplotlib.pyplot as plt
	if clear:
		plt.clf()

	plt.plot(r, ps['sb,' + band], linewidth=2, **kwargs)
	if axes:
		plt.set_ylim(max(ps['sb,' + band]), min(ps['sb,' + band]))
	else:
		plt.ylim(max(ps['sb,' + band]), min(ps['sb,' + band]))
	if fit_exp:
		exp_inds = np.where(r.in_units('kpc') > fit_exp)
		expfit = np.polyfit(np.array(r[exp_inds]),
							np.array(ps['sb,' + band][exp_inds]), 1)

		# 1.0857 is how many magnitudes a 1/e decrease is
		print(("h: ", 1.0857 / expfit[0], "  u_0:", expfit[1]))

		fit = np.poly1d(expfit)
		if 'label' in kwargs:
			del kwargs['label']
		if 'linestyle' in kwargs:
			del kwargs['linestyle']
		plt.plot(r, fit(r), linestyle='dashed', **kwargs)
	if fit_sersic:
		sersic_inds = np.where(r.in_units('kpc') < fit_sersic)
		sersicfit = np.polyfit(np.log10(np.array(r[sersic_inds])),
							   np.array(ps['sb,' + band][sersic_inds]), 1)
		fit = np.poly1d(sersicfit)
		print(("n: ", sersicfit[0], "  other: ", sersicfit[1]))
		if 'label' in kwargs:
			del kwargs['label']
		if 'linestyle' in kwargs:
			del kwargs['linestyle']
		plt.plot(r, fit(r), linestyle='dashed', **kwargs)
		#import pdb; pdb.set_trace()
	if axes:
		if print_ylabel:
			plt.set_ylabel(band + '-band Surface brightness [mag as$^{-2}$]')
	else:
		plt.xlabel('R [kpc]')
		plt.ylabel(band + '-band Surface brightness [mag as$^{-2}$]')
	if filename:
		logger.info("Saving %s", filename)
		plt.savefig(filename)


def f(x, alpha, delta, g):
	return -np.log10(10.0 ** (x * alpha) + 1.0) + delta * (np.log10(1 + np.exp(x))) ** g / (1 + np.exp(10 ** -x))


def behroozi(xmasses, z, alpha=-1.412, Kravtsov=False):
	'''Based on Behroozi+ (2013) return what stellar mass corresponds to the
	halo mass passed in.

	**Usage**

	   >>> from pynbody.plot.stars import moster
	   >>> xmasses = np.logspace(np.log10(min(totmasshalos)),1+np.log10(max(totmasshalos)),20)
	   >>> ystarmasses, errors = moster(xmasses,halo_catalog._halos[1].properties['z'])
	   >>> plt.fill_between(xmasses,np.array(ystarmasses)/np.array(errors),
						 y2=np.array(ystarmasses)*np.array(errors),
						 facecolor='#BBBBBB',color='#BBBBBB')
	'''
	loghm = np.log10(xmasses)
	# from Behroozi et al (2013)
	if Kravtsov: EPS = -1.642
	else: EPS = -1.777
	EPSpe = 0.133
	EPSme = 0.146

	EPSanu = -0.006
	EPSanupe = 0.113
	EPSanume = 0.361

	EPSznu = 0
	EPSznupe = 0.003
	EPSznume = 0.104

	EPSa = 0.119
	EPSape = 0.061
	EPSame = -0.012

	M1 = 11.514
	M1pe = 0.053
	M1me = 0.009

	M1a = -1.793
	M1ape = 0.315
	M1ame = 0.330

	M1z = -0.251
	M1zpe = 0.012
	M1zme = 0.125

	if Kravtsov: alpha=-1.779
	AL = alpha
	ALpe = 0.02
	ALme = 0.105

	ALa = 0.731
	ALape = 0.344
	ALame = 0.296

	if Kravtsov: DEL=4.394
	else: DEL = 3.508
	DELpe = 0.087
	DELme = 0.369

	DELa = 2.608
	DELape = 2.446
	DELame = 1.261

	DELz = -0.043
	DELzpe = 0.958
	DELzme = 0.071

	if Kravtsov: G=0.547
	else: G = 0.316
	Gpe = 0.076
	Gme = 0.012

	Ga = 1.319
	Gape = 0.584
	Game = 0.505

	Gz = 0.279
	Gzpe = 0.256
	Gzme = 0.081

	a = 1.0 / (z + 1.0)
	nu = np.exp(-4 * a ** 2)
	logm1 = M1 + nu * (M1a * (a - 1.0) + M1z * z)
	logeps = EPS + nu * (EPSanu * (a - 1.0) + EPSznu * z) - EPSa * (a - 1.0)
	analpha = AL + nu * ALa * (a - 1.0)
	delta = DEL + nu * DELa * (a - 1.0)
	g = G + nu * (Ga * (a - 1.0) + z * Gz)

	x = loghm - logm1
	f0 = -np.log10(2.0) + delta * np.log10(2.0) ** g / (1.0 + np.exp(1))
	smp = logm1 + logeps + f(x, analpha, delta, g) - f0

	if isinstance(smp, np.ndarray):
		scatter = np.zeros(len(smp))
	scatter = 0.218 - 0.023 * (a - 1.0)

	return 10 ** smp, 10 ** scatter


def moster(xmasses, z):
	'''Based on Moster+ (2013) return what stellar mass corresponds to the
	halo mass passed in.

	**Usage**

	   >>> from pynbody.plot.stars import moster
	   >>> xmasses = np.logspace(np.log10(min(totmasshalos)),1+np.log10(max(totmasshalos)),20)
	   >>> ystarmasses, errors = moster(xmasses,halo_catalog._halos[1].properties['z'])
	   >>> plt.fill_between(xmasses,np.array(ystarmasses)/np.array(errors),
						 y2=np.array(ystarmasses)*np.array(errors),
						 facecolor='#BBBBBB',color='#BBBBBB')
	'''
	hmp = np.log10(xmasses)
	# from Moster et al (2013)
	M10a = 11.590470
	M11a = 1.194913
	R10a = 0.035113
	R11a = -0.024729
	B10a = 1.376177
	B11a = -0.825820
	G10a = 0.608170
	G11a = 0.329275

	M10e = 0.236067
	M11e = 0.353477
	R10e = 0.00577173
	R11e = 0.00693815
	B10e = 0.153
	B11e = 0.225
	G10e = 0.059
	G11e = 0.173

	a = 1.0 / (z + 1.0)
	m1 = M10a + M11a * (1.0 - a)
	r = R10a + R11a * (1.0 - a)
	b = B10a + B11a * (1.0 - a)
	g = G10a + G11a * (1.0 - a)
	smp = hmp + np.log10(2.0 * r) - np.log10((10.0 ** (hmp - m1)) ** (-b) + (10.0 ** (hmp - m1)) **
											 (g))
	eta = np.exp(np.log(10.) * (hmp - m1))
	alpha = eta ** (-b) + eta ** g
	dmdm10 = (g * eta ** g + b * eta ** (-b)) / alpha
	dmdm11 = (g * eta ** g + b * eta ** (-b)) / alpha * (1.0 - a)
	dmdr10 = np.log10(np.exp(1.0)) / r
	dmdr11 = np.log10(np.exp(1.0)) / r * (1.0 - a)
	dmdb10 = np.log10(np.exp(1.0)) / alpha * np.log(eta) * eta ** (-b)
	dmdb11 = np.log10(np.exp(1.0)) / alpha * \
		np.log(eta) * eta ** (-b) * (1.0 - a)
	dmdg10 = -np.log10(np.exp(1.0)) / alpha * np.log(eta) * eta ** g
	dmdg11 = -np.log10(np.exp(1.0)) / alpha * \
		np.log(eta) * eta ** g * (1.0 - a)
	sigma = np.sqrt(dmdm10 * dmdm10 * M10e * M10e + dmdm11 * dmdm11 * M11e * M11e + dmdr10 * dmdr10 * R10e * R10e + dmdr11 * dmdr11 * R11e * R11e + dmdb10 * dmdb10 * B10e * B10e + dmdb11 * dmdb11 * B11e
					* B11e + dmdg10 * dmdg10 * G10e * G10e + dmdg11 * dmdg11 * G11e * G11e)
	return 10 ** smp, 10 ** sigma

def hudson(xmasses, z):
	''' Based on Hudson+ (2014), returns what stellar mass corresponds to the
	halo mass passed in.  This is the only SMHMR function that is not based
	on abundance matching, but instead uses date from CFHTLenS galaxy lensing data.
	   >>> from pynbody.plot.stars import hudson
	   >>> xmasses = np.logspace(np.log10(min(totmasshalos)),1+np.log10(max(totmasshalos)),20)
	   >>> ystarmasses, errors = hudson(xmasses,halo_catalog._halos[1].properties['z'])
	   >>> plt.fill_between(xmasses,np.array(ystarmasses)/np.array(errors),
	                     y2=np.array(ystarmasses)*np.array(errors),
	                     facecolor='#BBBBBB',color='#BBBBBB')
	'''
	f05 = 0.0414
	f05_err = 0.0024
	fz = 0.029
	fz_err = 0.009
	log10M05 = 12.07
	log10M05_err = 0.07
	Mz = 0.09
	Mz_err = 0.24
	beta = 0.69
	beta_err = 0.09
	gamma = 0.8

	f1 = f05 + (z-0.5)*fz
	f1_err = np.sqrt(f05_err**2+(z-0.5)**2*fz_err**2)

	M1_exp = (log10M05+(z-0.5)*Mz)
	M1_exp_err = np.sqrt(log10M05_err**2+(z-0.5)**2*Mz_err**2)

	M1 = np.power(10, M1_exp)
	M1_err = np.abs(M1*np.log(10)*M1_exp_err)

	beta_term = np.power(xmasses/M1, -beta)
	beta_term_err = np.sqrt((beta/M1*M1_err/M1**2)**2 + (np.log(M1)*beta_err)**2)

	gamma_term = np.power(xmasses/M1, gamma)
	gamma_term_err = np.abs(gamma_term*gamma*M1_err/M1**3)

	fstar_denom = beta_term + gamma_term
	fstar_denom_err = np.sqrt(beta_term*2+ gamma_term**2)

	fstar = 2.0*f1/fstar_denom
	fstar_err = np.sqrt((f1_err/f1)**2 + (fstar_denom_err/fstar_denom)**2)
	return fstar*xmasses, 2.0/fstar_err

def subfindguo(halo_catalog, clear=False, compare=True, baryfrac=False,
		filename=False, **kwargs):
	'''Stellar Mass vs. Halo Mass

	Takes a halo catalogue and plots the member stellar masses as a
	function of halo mass.

	Usage:

	>>> import pynbody.plot as pp
	>>> h = s.halos()
	>>> pp.guo(h,marker='+',markerfacecolor='k')

	**Options:**

	*compare* (True): Should comparison line be plotted?
		 If compare = 'guo', Guo+ (2010) plotted instead of Behroozi+ (2013)

	*baryfrac* (False):  Should line be drawn for cosmic baryon fraction?

	*filename* (None): name of file to which to save output
	'''

	# if 'marker' not in kwargs :
	#    kwargs['marker']='o'

	starmasshalos = []
	totmasshalos = []
	f_b = halo_catalog[0].properties['omegaB0']/halo_catalog[0].properties['omegaM0']
	for halo in halo_catalog:
		for subhalo in halo.sub:
			subhalo.properties['MassType'].convert_units('Msol')
			halostarmass = subhalo.properties['MassType'][4]
			if halostarmass:
				starmasshalos.append(halostarmass)
				totmasshalos.append(np.sum(subhalo.properties['MassType']))

	if clear:
		plt.clf()

	plt.loglog(totmasshalos, starmasshalos, 'o', **kwargs)
	plt.xlabel('Total Halo Mass')
	plt.ylabel('Halo Stellar Mass')

	if compare:
		xmasses = np.logspace(
			np.log10(min(totmasshalos)), 1 + np.log10(max(totmasshalos)), 20)
		if compare == 'guo':
			# from Sawala et al (2011) + Guo et al (2009)
			ystarmasses = xmasses*0.129*((xmasses/2.5e11)**-0.926 + (xmasses/2.5e11)**0.261)**-2.44
		else :
			ystarmasses, errors = behroozi(xmasses,halo_catalog._halos[1].properties['z'])
		plt.fill_between(xmasses,np.array(ystarmasses)/np.array(errors),
						 y2=np.array(ystarmasses)*np.array(errors),
						 facecolor='#BBBBBB',color='#BBBBBB')
		plt.loglog(xmasses,ystarmasses,label='Behroozi et al (2013)')

	if baryfrac :
		xmasses = np.logspace(np.log10(min(totmasshalos)),1+np.log10(max(totmasshalos)),20)
		ystarmasses = xmasses*f_b
		plt.loglog(xmasses,ystarmasses,linestyle='dotted',label='f_b = '+'%.2f' % f_b)
		ystarmasses = xmasses*0.1*f_b
		plt.loglog(xmasses,ystarmasses,linestyle='dashed',label='0.1 f_b = '+'%.2f' % (0.1*f_b))

	plt.axis([0.8*min(totmasshalos),1.2*max(totmasshalos),
			  0.8*min(starmasshalos),1.2*max(starmasshalos)])

	if (filename):
		logger.info("Saving %s", filename)
		plt.savefig(filename)

def guo(halo_catalog, clear=False, compare=True, baryfrac=False,
		filename=False, **kwargs):
	'''Stellar Mass vs. Halo Mass

	Takes a halo catalogue and plots the member stellar masses as a
	function of halo mass.

	Usage:

	>>> import pynbody.plot as pp
	>>> h = s.halos()
	>>> pp.guo(h,marker='+',markerfacecolor='k')

	**Options:**

	*compare* (True): Should comparison line be plotted?
		 If compare = 'guo', Guo+ (2010) plotted instead of Behroozi+ (2013)

	*baryfrac* (False):  Should line be drawn for cosmic baryon fraction?

	*filename* (None): name of file to which to save output
	'''

	# if 'marker' not in kwargs :
	#    kwargs['marker']='o'

	starmasshalos = []
	totmasshalos = []

	halo_catalog._halos[1]['mass'].convert_units('Msol')

	for i in np.arange(len(halo_catalog._halos)) + 1:
		halo = halo_catalog[i]
		halostarmass = np.sum(halo.star['mass'])
		if halostarmass:
			starmasshalos.append(halostarmass)
			totmasshalos.append(np.sum(halo['mass']))

	if clear:
		plt.clf()

	plt.loglog(totmasshalos, starmasshalos, 'o', **kwargs)
	plt.xlabel('Total Halo Mass')
	plt.ylabel('Halo Stellar Mass')

	if compare:
		xmasses = np.logspace(
			np.log10(min(totmasshalos)), 1 + np.log10(max(totmasshalos)), 20)
		if compare == 'guo':
			# from Sawala et al (2011) + Guo et al (2009)
			ystarmasses = xmasses*0.129*((xmasses/2.5e11)**-0.926 + (xmasses/2.5e11)**0.261)**-2.44
		else :
			ystarmasses, errors = behroozi(xmasses,halo_catalog._halos[1].properties['z'])
		plt.fill_between(xmasses,np.array(ystarmasses)/np.array(errors),
						 y2=np.array(ystarmasses)*np.array(errors),
						 facecolor='#BBBBBB',color='#BBBBBB')
		plt.loglog(xmasses,ystarmasses,label='Behroozi et al (2013)')

	if baryfrac :
		xmasses = np.logspace(np.log10(min(totmasshalos)),1+np.log10(max(totmasshalos)),20)
		ystarmasses = xmasses*0.04/0.24
		plt.loglog(xmasses,ystarmasses,linestyle='dotted',label='f_b = 0.16')

	plt.axis([0.8*min(totmasshalos),1.2*max(totmasshalos),
			  0.8*min(starmasshalos),1.2*max(starmasshalos)])

	if (filename):
		logger.info("Saving %s", filename)
		plt.savefig(filename)



"""
Tools for connecting different simulation snaphots

The bridge module allows you to take a subview (e.g. a halo) from one snapshot and 'push' it into the other. This is
especially useful if the two snapshots are different time outputs of the same simulation, or closely-related
simulations (e.g. one DMO and one hydro).

Once connected, bridge called on a specific subset of particles in output1 will trace these particles back (or
forward) to the output2, enabling observing a change in their properties, such as position, temperature, etc.

For an introduction on how to use bridges, see the :doc:`tutorial </tutorials/bridge>`.

"""

from __future__ import annotations

import abc
import typing
import weakref

import numpy as np

from .. import util
from . import _bridge

if typing.TYPE_CHECKING:
    from .. import snapshot

class AbstractBridge(abc.ABC):
    """The abstract base class for bridges between two snapshots.

    For more information see the module documentation for :mod:`pynbody.bridge`, or the
    :doc:`bridge tutorial </tutorials/bridge>`.
    """

    def __init__(self, start: snapshot.SimSnap, end: snapshot.SimSnap):
        self._start = weakref.ref(start)
        self._end = weakref.ref(end)

    @abc.abstractmethod
    def __call__(self, s: snapshot.SubSnap) -> snapshot.SubSnap:
        """Map from a ``SubSnap`` at one end of the bridge, to the corresponding ``SubSnap`` at the other end"""


    def _get_ends(self):
        start = self._start()
        end = self._end()
        if start is None or end is None:
            raise RuntimeError("Stale reference to start or endpoint")
        return start, end


    def match_halos(self, halos_1, halos_2, /, threshold=0.5, use_family=None, fill_value=-1, use_halo_indexes=False):
        """Given halo catalogues, identify the likely halo number in the second catalogue for each halo in the first.

        For example, if a bridge ``b`` links snapshot ``f1`` (high redshift) to ``f2`` (low redshift) and we perform:

        >>> h1 = f1.halos()
        >>> h2 = f2.halos()
        >>> cat = b.match_halos(h1, h2)

        then ``cat`` is now a dictionary such that f1.halos()[i] is the major progenitor for``f2.halos()[cat[i]]``,
        assuming ``cat[i]`` is positive.

        Halos which cannot be matched because they have too few particles in common give the result ``fill_value``,
        default -1. This is determined by the given threshold fraction of particles in common (specified by
        ``threshold``, default 50%)

        Parameters
        ----------

        halos_1 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the first snapshot.

        halos_2 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the second snapshot.

        threshold : float
            The minimum fraction of particles in common for a match to be considered. Default is 0.5.

        use_family : str
            Only match particles of this family. Default is None, in which case all particles are matched.
            Setting this to a family name can be useful if matching between two different simulations where the
            relationship between DM particles is known, but perhaps the relationship between gas particles is not
            (e.g. a Ramses simulation where actually the gas 'particles' are cells)

        use_halo_indexes : bool
            If True, instead of returning a dictionary mapping to halo numbers, return a numpy array that matches
            halo IDs (which are zero-based indexes into the full list of halos; for more information see
            the nomenclature guide in the documentation for :class:`~pynbody.halo.HaloCatalogue`. The default is
            False, so that halo numbers are used throughout.

        fill_value : int
            The value to use for halos that cannot be matched. Default is -1.

        """

        self._check_compatible_halo_catalogues(halos_1, halos_2)

        particles_in_common_matrix = self.count_particles_in_common(halos_1, halos_2, use_family=use_family)

        highest_commonality_index = particles_in_common_matrix.argmax(axis=1)
        highest_commonality = particles_in_common_matrix[np.arange(particles_in_common_matrix.shape[0]),
                                                         highest_commonality_index]
        # return nan for zero division
        with np.errstate(divide='ignore', invalid='ignore'):
            frac_commonality = highest_commonality/particles_in_common_matrix.sum(axis=1)
        frac_commonality[~np.isfinite(frac_commonality)] = 0

        invalid_matches = frac_commonality<threshold

        if use_halo_indexes:
            result = highest_commonality_index
            result[invalid_matches] = fill_value
            return result
        else:
            mapper_1 = halos_1.number_mapper
            mapper_2 = halos_2.number_mapper

            sources = mapper_1.index_to_number(np.arange(len(halos_1)))
            destinations = mapper_2.index_to_number(highest_commonality_index)
            destinations[invalid_matches] = fill_value

            return dict(zip(sources, destinations))

    def _check_compatible_halo_catalogues(self, halos_1, halos_2):
        if halos_1.base.ancestor is not self._start().ancestor or halos_2.base.ancestor is not self._end().ancestor:
            raise ValueError("The HaloCatalogues must be based on the correct snapshots")

    def fuzzy_match_halos(self, halos_1, halos_2, /, threshold=0.01, use_family=None,
                          use_halo_indexes=False) -> dict[int, list[tuple[int, float]]]:
        """Given halo catalogues, return possible matches within the second catalogue for each halo in the first.

        For details about the parameters to this function, see the documentation for :func:`match_catalog`.

        Parameters
        ----------

        halos_1 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the source snapshot

        halos_2 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the destination snapshot

        threshold : float
            The minimum fraction of particles in common for a match to be considered. Default is 0.01.

        use_family : str
            Only match particles of this family. Default is None, in which case all particles are matched.

        use_halo_indexes : bool
            If True, instead of returning halo numbers, return halo indexes. Default is False.

        Returns
        -------
        dict[int, list[tuple[int, float]]]
            Maps from halo number in the first catalogue to a list of tuples. Each tuple contains the halo number in the
            second catalogue and the fraction of particles in the first halo that are in the second halo. If
            ``use_halo_indexes`` is True, the halo numbers are replaced by halo indexes.
        """

        self._check_compatible_halo_catalogues(halos_1, halos_2)

        def map_index_to_output(index, for_halos):
            if use_halo_indexes:
                return index
            else:
                return for_halos.number_mapper.index_to_number(index)

        particles_in_common_matrix = self.count_particles_in_common(halos_1, halos_2, use_family=use_family)

        output = {}

        for source_index, row in enumerate(particles_in_common_matrix):
            if source_index >= len(halos_1):
                # count_particles_in_common returns a square matrix, so we might run over the end of the first catalogue
                break

            this_row_matches = []
            row_sum = row.sum()
            if row_sum > 0:
                frac_particles_transferred = row / row_sum
                above_threshold = np.where(frac_particles_transferred > threshold)[0]
                above_threshold = above_threshold[np.argsort(frac_particles_transferred[above_threshold])[::-1]]
                for column in above_threshold:
                    this_row_matches.append((map_index_to_output(column, halos_2), frac_particles_transferred[column]))

            output[map_index_to_output(source_index, halos_1)] = this_row_matches

        return output

    def count_particles_in_common(self, halos_1, halos_2, /, max_num_halos=None, use_family=None) -> np.ndarray:
        """Return a matrix with the number of particles transferred from ``groups_1`` to groups_2.

        Normally, :func:`match_catalog` (or :func:`fuzzy_match_catalog`) are easier to use, but this routine
        provides the maximal information.

        .. warning::
            This routine returns results in terms of halo indexes, rather than halo numbers. This is because halo
            indexes are guaranteed to be continuous starting at zero, while halo numbers may have gaps. If you need
            to convert from halo indexes to halo numbers, you can use :meth:`~pynbody.halo.HaloCatalogue.number_mapper`
            attribute of :class:`~pynbody.halo.HaloCatalogue`. Alternatively, use the :func:`match_catalog` method
            which returns results in terms of halo numbers.

        Parameters
        ----------
        halos_1 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the first snapshot (the one at the start of the bridge).
        halos_2 : pynbody.halo.HaloCatalogue
            The HaloCatalogue for the second snapshot (the one at the end of the bridge).
        max_num_halos : int, optional
            The maximum number of halos
        use_family : str
            Only match particles of this family. Default is None, in which case all particles are matched.

        Returns
        -------
        numpy.ndarray
            A matrix with the number of particles transferred from each halo in the first catalogue to each halo in the
            second. The size of the matrix is determined by the maximum number of halos in either catalogue, or
            by the value of ``max_num_halos`` if specified.

        """

        self._check_compatible_halo_catalogues(halos_1, halos_2)

        start, end = self._get_ends()

        if use_family:
            end = end[use_family]
            start = start[use_family]

        restricted_start_particles = self(self(start)) # map back and forth to get only particles that are held in common
        restricted_end_particles = self(restricted_start_particles) # map back to start in case a reordering is required

        restriction_end_indices = restricted_end_particles.get_index_list(end.ancestor)
        restriction_start_indices = restricted_start_particles.get_index_list(start.ancestor)

        assert len(restriction_end_indices) == len(restriction_start_indices), \
            ("Internal consistency failure in catalog_transfer_matrix: particles supposedly "
             "common to both simulations have two different lengths")


        # Need to account for the fact that get_group_array(only_family) will return an array of length Nfamily
        # while restriction_start_indices are global indeces to the ancestor snapshot,
        # which can start anywhere depending on the family ordering ==> Need to offset them back to start at 0.
        restriction_start_indices -= start.ancestor._get_family_slice(use_family).start
        restriction_end_indices -= end.ancestor._get_family_slice(use_family).start

        g1 = halos_1.get_group_array(family=use_family, use_index=True)[restriction_start_indices]
        g2 = halos_2.get_group_array(family=use_family, use_index=True)[restriction_end_indices]

        if max_num_halos is None:
            max_num_halos = max(len(halos_1), len(halos_2))

        transfer_matrix = _bridge.match(g1, g2, 0, max_num_halos)

        return transfer_matrix



    @util.deprecated("match_catalog is deprecated; use match_halos instead")
    def match_catalog(self, min_index=1, max_index=30, threshold=0.5, groups_1 = None, groups_2 = None,
                      use_family = None):
        """Deprecated alternative to :func:`match_halos`. Use that method instead."""
        fuzzy_matches = self.fuzzy_match_catalog(min_index, max_index, threshold, groups_1, groups_2, use_family)

        identification = np.zeros(max_index+1,dtype=int)

        for i,row in enumerate(fuzzy_matches):
            if len(row)>0:
                identification[i] = row[0][0]
            elif i<min_index:
                identification[i] = -2
            else:
                identification[i] = -1

        return identification


    @util.deprecated("fuzzy_match_catalog is deprecated; use fuzzy_match_halos instead")
    def fuzzy_match_catalog(self, min_index=1, max_index=30, threshold=0.01,
                            groups_1 = None, groups_2 = None, use_family=None, only_family=None):
        """Deprecated alternative to :func:`fuzzy_match_halos`. Use that method instead."""

        transfer_matrix = self.catalog_transfer_matrix(min_index,max_index,groups_1,groups_2,use_family,only_family)

        output = [[]]*min_index
        for row in transfer_matrix:
            this_row_matches = []
            if row.sum()>0:
                frac_particles_transferred = np.array(row,dtype=float)/row.sum()
                above_threshold = np.where(frac_particles_transferred>threshold)[0]
                above_threshold = above_threshold[np.argsort(frac_particles_transferred[above_threshold])[::-1]]
                for column in above_threshold:
                    this_row_matches.append((column+min_index, frac_particles_transferred[column]))

            output.append(this_row_matches)

        return output

    @util.deprecated("catalog_transfer_matrix is deprecated; use count_particles_in_common instead")
    def catalog_transfer_matrix(self, min_index=1, max_index=30, groups_1=None, groups_2=None,
                                use_family=None, only_family=None):
        """Deprecated interface to :func:`count_particles_in_common`. Use that method instead."""

        # the 'index' naming of the parameter is misleading -- in pynbody v1, the distinction between number and
        # index had not been drawn. In fact, by pynbody v2 nomenclature, this function accepts halo numbers, so
        # we rename internally to prevent confusion in the compatibility code below.

        min_number = min_index
        max_number = max_index

        del min_index, max_index

        if only_family is not None:
            if use_family!=only_family and use_family is not None:
                raise ValueError("use_family and only_family must be the same if both specified")
            use_family = only_family

        start, end = self._get_ends()
        if groups_1 is None:
            groups_1 = start.halos()
        else:
            assert groups_1.base.ancestor is start.ancestor

        if groups_2 is None:
            groups_2 = end.halos()
        else:
            assert groups_2.base.ancestor is end.ancestor

        max_number_1 = min(max_number, max(groups_1.keys()))
        max_number_2 = min(max_number, max(groups_2.keys()))

        indices_to_use_1 = groups_1.number_mapper.number_to_index(np.arange(min_number, max_number_1+1))
        indices_to_use_2 = groups_2.number_mapper.number_to_index(np.arange(min_number, max_number_2+1))

        max_index = max(indices_to_use_1.max(), indices_to_use_2.max())

        transfer_matrix = self.count_particles_in_common(groups_1, groups_2,
                                                         max_num_halos=max_index, use_family=use_family)

        transfer_matrix_restricted = transfer_matrix[indices_to_use_1, :][:, indices_to_use_2]

        return transfer_matrix_restricted

class OneToOneBridge(AbstractBridge):
    """Connects two snapshots with identical particle numbers and file layout.

    Particle ``i`` in the start point is the same as particle ``i`` in the end point.
    """

    def __init__(self, start: snapshot.SimSnap, end: snapshot.SimSnap):
        if len(start) != len(end):
            raise ValueError("OneToOneBridge requires snapshots of the same length")
        super().__init__(start, end)

    def __call__(self, s):
        start, end = self._get_ends()

        if s.is_descendant(start):
            return end[s.get_index_list(start)]
        elif s.is_descendant(end):
            return start[s.get_index_list(end)]
        else:
            raise RuntimeError("Not a subview of either end of the bridge")

class OrderBridge(AbstractBridge):
    """Connects to snapshots that both have arrays of identity integers (``iord`` or similar) to identify particles.

    Particles ``i_start`` and ``i_end`` are defined to be the same if and only if
    ``start[order_array][i_start] == start[order_array][i_end]``.
    """

    def __init__(self, start, end, order_array="iord", monotonic=True, allow_family_change=False):
        """Initialise the ``OrderBridge``

        Parameters
        ----------

        start : snapshot.SimSnap
            The start point of the bridge

        end : snapshot.SimSnap
            The end point of the bridge

        order_array : str
            The name of the array that is used to identify particles. Default is ``iord``.

        monotonic : bool
            If ``True``, the order_array must be monotonically increasing in both ends of the bridge.
            Note that this is not checked for you. If ``False``, the bridging is slower but this is the failsafe option.

        allow_family_change : bool
            If ``True``, the bridge will allow particles to change family going from one end to the other of the
            bridge. Otherwise, it is assumed that the family of a particle is conserved.

        """
        super().__init__(start, end)
        self._order_array = order_array
        self.monotonic = monotonic
        self.allow_family_change = allow_family_change

    def __call__(self, s):

        start, end = self._get_ends()

        if s.is_descendant(start):
            from_ = start
            to_ = end

        elif s.is_descendant(end):
            from_ = end
            to_ = start

        else:
            raise RuntimeError("Not a subview of either end of the bridge")

        iord_to = np.asarray(to_[self._order_array]).view(np.ndarray)
        iord_from = np.asarray(s[self._order_array]).view(np.ndarray)

        if not self.monotonic:
            iord_map_to = np.argsort(iord_to)
            iord_map_from = np.argsort(iord_from)
            iord_to = iord_to[iord_map_to]
            iord_from = iord_from[iord_map_from]

        output_index, found_match = _bridge.bridge(iord_to, iord_from)

        if not self.monotonic:
            output_index = iord_map_to[output_index[np.argsort(iord_map_from)][found_match]]
        else:
            output_index = output_index[found_match]

        if self.allow_family_change:
            new_family_index = to_._family_index()[output_index]

            # stable sort by family:
            output_index = output_index[np.lexsort((new_family_index,))]

        return to_[output_index]


def bridge_factory(a: snapshot.SimSnap, b: snapshot.SimSnap) -> AbstractBridge:
    """Create a bridge connecting the two specified snapshots.

    This function will determine the best type of :ref:`Bridge` to construct between the two snapshots, and return it.
    It is called by :func:`pynbody.snapshot.simsnap.SimSnap.bridge`.

    For more information see :doc:`the bridge tutorial </tutorials/bridge>`.
    """

    not_sure_error = "Don't know how to automatically bridge between two simulations of different formats. "\
                     "You will need to create your bridge manually by instantiating either the OneToOneBridge or "\
                     "OrderBridge class appropriately."

    from ..snapshot import gadget, gadgethdf, nchilada, ramses, tipsy
    a_top = a.ancestor
    b_top = b.ancestor

    if type(a_top) is not type(b_top):
        raise RuntimeError(not_sure_error)

    if (isinstance(a_top, tipsy.TipsySnap) or isinstance(a_top, nchilada.NchiladaSnap)):
        if "iord" in a_top.loadable_keys():
            return OrderBridge(a_top, b_top, monotonic=True)
        else:
            return OneToOneBridge(a_top, b_top)
    elif isinstance(a_top, gadget.GadgetSnap) or isinstance(a_top, gadgethdf.GadgetHDFSnap):
        return OrderBridge(a_top, b_top, monotonic=False, allow_family_change=True)
    elif isinstance(a_top, ramses.RamsesSnap):
        if len(a.gas) > 0 or len(b.gas) > 0:
            raise RuntimeError("Cannot bridge AMR gas cells")
        return OrderBridge(a_top, b_top, monotonic=False)
    else:
        raise RuntimeError(not_sure_error)



"""Tool for mapping from individual simulation code's naming conventions to pynbody naming conventions.

This module is a detail that is not intended to be used directly by end users. It is used by the snapshot classes to
map between the names of arrays in different simulation codes.
"""

import configparser
import sys

from .. import config_parser


def setup_name_maps(config_name, gadget_blocks=False, with_alternates=False):
    name_map = {}
    rev_name_map = {}
    alternates = {}
    try:
        for a, b in config_parser.items(config_name):
            if sys.version_info[0] == 2:
                if gadget_blocks:
                    a = a.upper().ljust(4)
            else:
                if gadget_blocks:
                    a = a.upper().ljust(4).encode("utf-8")
            if b in name_map:
                alternates[name_map[b]] = b
            rev_name_map[a] = b
            name_map[b] = a
    except configparser.NoOptionError:
        pass

    if with_alternates:
        return name_map, rev_name_map, alternates
    else:
        return name_map, rev_name_map


def name_map_function(name_map, rev_name_map):
    def _translate_array_name(name, reverse=False):
        try:
            if reverse:
                return rev_name_map[name]
            else:
                return name_map[name]
        except KeyError:
            return name

    return _translate_array_name


class AdaptiveNameMapper:
    """A class to map between the names of arrays in different simulation codes.

    This class is designed to be used in a context where the names of arrays in a simulation code may not be fully
    known in advance. For example, we might be unsure whether pynbody's ``pos`` array corresponds to the ``Coordinates``
    array or the ``Position`` array in a given simulation snapshot. This class allows us possible different mappings
    to be given in the configuration, and then once one of the mappings is used, it is locked in for the duration of
    the object's lifetime.

    """
    def __init__(self, config_name, gadget_blocks=False):
        self._name_map, self._rev_name_map, self._alternates = setup_name_maps(config_name, gadget_blocks,True)

    def _select_alternate_target(self,name):
        self._name_map[self._alternates[name]]=name
        self._rev_name_map[name]=self._alternates[name]

    def _select_alternate_target_if_required(self,name):
        if name not in list(self._name_map.values()) and name in self._alternates:
            self._select_alternate_target(name)

    def __call__(self, name, reverse=False):
        if reverse:
            self._select_alternate_target_if_required(name)
        current_map = self._rev_name_map if reverse else self._name_map
        return current_map.get(name, name)



import copy
import gc

import numpy as np
import numpy.testing as npt
import pytest

import pynbody


@pytest.fixture
def test_simulation_with_copy():
    s = pynbody.new(dm=1000)
    s['pos'] = np.random.normal(scale=1.0, size=s['pos'].shape)
    s['vel'] = np.random.normal(scale=1.0, size=s['vel'].shape)
    s['mass'] = np.random.uniform(1.0, 10.0, size=s['mass'].shape)
    return s, copy.deepcopy(s)


def test_translate(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with f.translate([1, 0, 0]):
        npt.assert_almost_equal(f['pos'], original['pos'] + [1, 0, 0])

    # check moved back
    npt.assert_almost_equal(f['pos'], original['pos'])

    # try again with with abnormal exit
    try:
        with f.translate([1, 0, 0]):
            npt.assert_almost_equal(f['pos'], original['pos'] + [1, 0, 0])
            raise RuntimeError
    except RuntimeError:
        pass

    npt.assert_almost_equal(f['pos'], original['pos'])


def test_v_translate(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with f.offset_velocity([1, 0, 0]):
        npt.assert_almost_equal(f['vel'], original['vel'] + [1, 0, 0])

    # check moved back
    npt.assert_almost_equal(f['vel'], original['vel'])

    # try again with with abnormal exit
    try:
        with f.offset_velocity([1, 0, 0]):
            npt.assert_almost_equal(f['vel'], original['vel'] + [1, 0, 0])
            raise RuntimeError
    except RuntimeError:
        pass

    npt.assert_almost_equal(f['vel'], original['vel'])


def test_vp_translate(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with f.translate([1, 0, 0]).offset_velocity([2,0,0]):
        npt.assert_almost_equal(f['vel'], original['vel'] + [2, 0, 0])
        npt.assert_almost_equal(f['pos'], original['pos'] + [1, 0, 0])

    # check moved back
    npt.assert_almost_equal(f['vel'], original['vel'])
    npt.assert_almost_equal(f['pos'], original['pos'])

    # try again with with abnormal exit
    try:
        with f.translate([1, 0, 0]).offset_velocity([2,0,0]):
            npt.assert_almost_equal(f['vel'], original['vel'] + [2, 0, 0])
            npt.assert_almost_equal(f['pos'], original['pos'] + [1, 0, 0])
            raise RuntimeError
    except RuntimeError:
        pass

    npt.assert_almost_equal(f['vel'], original['vel'])
    npt.assert_almost_equal(f['pos'], original['pos'])


def test_rotate(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with f.rotate_x(90):
        npt.assert_almost_equal(f['y'], -original['z'])
        npt.assert_almost_equal(f['z'], original['y'])

    npt.assert_almost_equal(f['pos'], f['pos'])

def test_family_rotate():
    """Test that rotating a family works as expected

    See issue #728"""
    f = pynbody.new(dm=10, gas=20, bh=10)
    f['pos'] = np.zeros((40, 3))
    f['pos'][:, 0] = 1.0

    del f['vel']

    f.bh['vel'] = np.zeros((10, 3))
    f.bh['vel'][:, 0] = 1.0

    f.rotate_z(90)

    npt.assert_almost_equal(f['pos'][:, 1], 1.0)
    npt.assert_almost_equal(f.bh['vel'][:, 1], 1.0)

def test_chaining(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with f.rotate_x(90).translate([0, 1, 0]):
        npt.assert_almost_equal(f['y'], 1.0 - original['z'])
        npt.assert_almost_equal(f['z'], original['y'])

    npt.assert_almost_equal(f['pos'], original['pos'])


def test_halo_managers(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with pynbody.analysis.angmom.sideon(f, disk_size=1.0):
        pass

    npt.assert_almost_equal(f['pos'], original['pos'])

def test_repr(test_simulation_with_copy):
    f, _ = test_simulation_with_copy

    with f.rotate_x(45).translate([0, 1, 0]).offset_velocity([0, 0, 1]) as tx:
        assert str(tx) == "rotate_x(45), translate, offset_velocity"
        assert repr(tx) == "<Transformation rotate_x(45), translate, offset_velocity>"

def test_null(test_simulation_with_copy):
    f, original = test_simulation_with_copy

    with pynbody.transformation.NullTransformation(f) as tx:
        npt.assert_almost_equal(f['pos'], original['pos'])
        assert tx.sim is not None

    with tx.rotate_x(90) as tx2:
        assert tx2.next_transformation is None
        assert str(tx2) == "rotate_x(90)"
        npt.assert_almost_equal(f['y'], -original['z'])


    npt.assert_almost_equal(f['pos'], original['pos'])

def test_weakref():
    f = pynbody.new(dm=10)

    tx1 = f.rotate_y(90)
    tx2 = tx1.rotate_x(90).translate([0, 1, 0])
    assert tx1.sim is not None
    assert tx2.sim is not None
    del f
    gc.collect()
    assert tx1.sim is None
    assert tx2.sim is None

@pytest.mark.parametrize("family", [True, False])
def test_derived_3d_array(family):
    """Test for a bug where transformations would try to rotate derived arrays, raising an error"""
    f = pynbody.new(dm=5, gas=5)

    @pynbody.derived_array
    def test_derived_3d_array_transformation(sim):
        return np.random.normal(size=(len(sim), 3))

    if family:
        f = f.dm

    _ = f['test_derived_3d_array_transformation']

    with f.ancestor.rotate_y(90):
        _ = f['test_derived_3d_array_transformation']



import sys

import numpy as np

cimport cython
cimport numpy as npc
from cython cimport integral

# The following slightly odd repetitiveness is to force Cython to generate
# code for different permutations of the possible integer inputs.
#
# Using just one fused type requires the types to be synchronized across all
# inputs.

ctypedef fused integral_1:
    npc.int32_t
    npc.uint32_t
    npc.int64_t
    npc.uint64_t

ctypedef fused integral_2:
    npc.int32_t
    npc.uint32_t
    npc.int64_t
    npc.uint64_t



@cython.boundscheck(False)
@cython.wraparound(False)
def bridge(npc.ndarray[integral_1, ndim=1] iord_to,
            npc.ndarray[integral_2, ndim=1] iord_from) :

    cdef npc.ndarray[npc.int64_t, ndim=1] output_index
    cdef npc.ndarray[npc.uint8_t, ndim=1] found_match
    cdef npc.int64_t i=0, i_to=0, j=0
    cdef npc.int64_t length = len(iord_from)
    cdef npc.int64_t length_to = len(iord_to)

    found_match = np.empty(length,dtype=np.uint8)
    output_index = np.empty(length,dtype=np.int64)
    for i in range(length) :
        i_from = iord_from[i]
        while i_from>iord_to[i_to] and i_to<length_to :
            i_to+=1
        if i_from==iord_to[i_to] :
            output_index[i] = i_to
            found_match[i] = 1 # true
        else:
            output_index[i] = 0
            found_match[i] = 0 # false

    return (output_index, found_match.astype(np.bool_))


@cython.boundscheck(False)
@cython.wraparound(False)
def match(npc.ndarray[integral_1, ndim=1] group_list_1,
          npc.ndarray[integral_2, ndim=1] group_list_2,
          int imin, int imax):
    cdef size_t i, l = len(group_list_1)
    cdef int g1, g2
    cdef npc.ndarray[npc.int64_t, ndim=2] output = np.zeros((imax+1-imin,imax+1-imin),dtype=np.int64)

    assert len(group_list_2)==l

    with nogil:
        for i in range(l):
            g1 = group_list_1[i]
            g2 = group_list_2[i]
            if g1<=imax and g2<=imax and g1>=imin and g2>=imin :
                output[g1-imin,g2-imin]+=1

    return output



"""
Routines for plotting smoothed quantities

"""

import warnings

import matplotlib
import numpy as np
import pylab as p

from .. import config, sph, units as _units


def _width_in_sim_units(sim, width):
	if isinstance(width, str) or issubclass(width.__class__, _units.UnitBase):
		if isinstance(width, str):
			width = _units.Unit(width)
		width = width.in_units(sim['pos'].units, **sim.conversion_context())
	return width

def sideon_image(sim, *args, **kwargs):
    """

    Rotate the simulation so that the disc of the passed halo is
    side-on, then make an SPH image by passing the parameters into
    the function image

    For a description of keyword arguments see :func:`~pynbody.plot.sph.image`.

    """

    from ..analysis import angmom

    with angmom.sideon(sim):
        return image(sim, *args, **kwargs)


def faceon_image(sim, *args, **kwargs):
    """

    Rotate the simulation so that the disc of the passed halo is
    face-on, then make an SPH image by passing the parameters into
    the function image

    For a description of keyword arguments see :func:`~pynbody.plot.sph.image`.

    """

    from ..analysis import angmom

    with angmom.faceon(sim):
        return image(sim, *args, **kwargs)


def contour(sim, qty, width="10 kpc", resolution=None, units=None, axes=None, label=True, log=True, weight=None,
			contour_kwargs=None, smooth_min=0.0, _transform=None):
    """Create an image of the given quantity then turn it into contours.

    Parameters
    ----------

    sim : pynbody.snapshot.simsnap.SimSnap
        The simulation snapshot to plot. The image is generated in the plane z=0, or
        projected along the z axis.

    qty : str | pynbody.array.SimArray
        The name of the array to interpolate. Default is 'rho', which gives a density
        image. Alternatively, an array can be passed in.

    width : str or float, optional
        The overall width and height of the plot. If a float, it is assumed
        to be in units of sim['pos']. It can also be passed in as a string indicating
        the units, e.g. '10 kpc'. (Default is '10 kpc')

    resolution : int, optional
        The number of pixels wide and tall. (Default is determined by the
        :ref:`configuration file <configuration>`.)

    units : str or pynbody.units.Unit, optional
        The units of the output. Default is None, in which case the units of the input
        quantity are used. If the units correspond to integrating the quantity along
        a spatial dimension, the output is a projected image. For example, if the units
        are 'Msol kpc^-2', and the quantity is 'rho', the output is a projected image of
        the surface density.

    axes : matplotlib.axes.Axes, optional
        Axes instance on which the image will be shown; if None, the current pyplot figure is
        used. (Default is False)

    label : bool, optional
        Whether to label the contours. (Default is True)

    log : bool, optional
        If True, the image is log-scaled before being contoured. (Default is True)

    weight : str, optional
        If set, the requested quantity is volume-averaged down the line of sight, weighted either
        by volume (if weight is True) or by a specified quantity. (Default is None)

    contour_kwargs : dict, optional
        Additional keyword arguments to pass to the matplotlib contour function. (Default is None)

    smooth_min : float, optional
        The minimum size of the smoothing kernel, either as a float or a unit string.
		Setting this to a non-zero value makes smoother, clearer contours but loses fine detail.
		Default is 0.0.

    _transform : function, optional
        A function to apply to the image before contouring. (Default is None)

    """

    if resolution is None:
        resolution = config['image-default-resolution']

    if axes is None:
        axes = p.gca()

    if contour_kwargs is None:
        contour_kwargs = {}

    width = _width_in_sim_units(sim, width)
    pixel_size = width / resolution

    # width of image must be a pixel wider than the width of the final contour field, since the contours
    # are based on the centres of the pixels not their edges
    im = image(sim.s, qty, width=width + pixel_size, weight=weight, units=units, log=log, clear=False, noplot=True,
               resolution=resolution, smooth_min=smooth_min)

    if _transform:
        im = _transform(im)

    # width of image was expanded above, so this is now the positions of the centres
    X = np.linspace(-width/2, width/2, resolution)

    CS = axes.contour(X, X, im, **contour_kwargs)
    if label:
        axes.clabel(CS, fontsize=12, inline=True)


def velocity_image(sim, qty='rho', vector_qty='vel', width="10 kpc", mode='quiver',
                   vector_color='black', vector_edgecolor='black', vector_resolution=40,
                   vector_scale=None, key=True, key_x=0.5, key_y=0.88,
                   key_color='k', key_edge_color='k', key_bg_color='w',
                   key_length="100 km s**-1",
                   stream_density=1.0, stream_linewidth = 1.0,
                   weight=None, restrict_depth=False,
                   **kwargs):
    """
    Make an SPH image of the given simulation with velocity vectors overlaid on top.

    Any keyword argument that can be passed to :func:`~pynbody.plot.sph.image` can also be passed
    to this function. See that function for a full list of options.

    Parameters
    ----------

    sim : pynbody.snapshot.simsnap.SimSnap
        The simulation snapshot to plot. The image is generated in the plane z=0, or
        projected along the z axis.

    qty : str | pynbody.array.SimArray
        The name of the array to interpolate. Default is 'rho', which gives a density
        image. Alternatively, an array can be passed in.

    vector_qty : str
        The name of the array to use for the vectors. Default is 'vel'.

    width : str or float, optional
        The overall width and height of the plot. If a float, it is assumed
        to be in units of sim['pos']. It can also be passed in as a string indicating
        the units, e.g. '10 kpc'. (Default is '10 kpc')

    mode : str, optional
        The type of plot to make. Options are 'quiver' or 'stream'. (Default is 'quiver')

    vector_color : str, optional
        The color of the velocity vectors. (Default is 'black')

    vector_edgecolor : str, optional
        The color for the edges of the velocity vectors. (Default is 'black')

    vector_resolution : int, optional
        The number of velocity vectors to generate in each dimension. (Default is 40)

    vector_scale : str or float, optional
        The length of a vector that would result in a displayed length of the
        figure width/height. This can be provided as a unit string or a float, in which
        case it is interpreted as having the same units as the vector array being plotted.
        Default is None, in which case the vectors are scaled by matplotlib's quiver function.
        This option is only used in 'quiver' mode.

    key : bool, optional
        Whether or not to inset a key showing the scale of the vectors.
        Only used if in 'quiver' mode. (Default is True)

    key_x : float, optional
        The x position of the key in 'quiver' mode, if key is True. (Default is 0.5)

    key_y : float, optional
        The y position of the key in 'quiver' mode, if key is True. (Default is 0.88)

    key_color : str, optional
        The color of the key arrow/text in 'quiver' mode, if key is True. (Default is 'white')

    key_edge_color : str, optional
        The color of the border around the key in 'quiver' mode, if key is True. (Default is 'black')

    key_bg_color : str, optional
        The color of the background of the key in 'quiver' mode, if key is True. (Default is 'white')

    key_length : str, optional
        The velocity to use for the key in 'quiver' mode, if key is True. (Default is '100 km s**-1')

    stream_density : float, optional
        The density of stream lines in 'stream' mode. (Default is 1.0)

    stream_linewidth : float, optional
        The width of stream lines in 'stream' mode. (Default is 1.0)

    **kwargs :
        Any additional keyword arguments to pass to :func:`~pynbody.plot.sph.image`.

    """



    if 'av_z' in kwargs:
        weight = kwargs.pop('av_z')
        warnings.warn("av_z is deprecated; use weight instead", DeprecationWarning)



    vx_name, vy_name, _ = sim._array_name_ND_to_1D(vector_qty)

    weight_for_vector = weight

    if 'units' in kwargs and _units_imply_projection(sim, qty, _units.Unit(kwargs['units'])) and weight is None:
        weight_for_vector = 'rho'

    vx = image(sim, qty=vx_name, width=width, log=False,
               resolution=vector_resolution, noplot=True, weight=weight_for_vector,
               restrict_depth=restrict_depth)
    vy = image(sim, qty=vy_name, width=width, log=False,
               resolution=vector_resolution, noplot=True, weight=weight_for_vector,
               restrict_depth=restrict_depth)
    key_unit = _units.Unit(key_length)

    if isinstance(width, str) or issubclass(width.__class__, _units.UnitBase):
        if isinstance(width, str):
            width = _units.Unit(width)
        width = width.in_units(sim['pos'].units, **sim.conversion_context())

    width = float(width)

    pixel_size = width / float(vector_resolution)
    X, Y = np.meshgrid(np.linspace(-width / 2 + pixel_size/2, width / 2 - pixel_size/2, vector_resolution),
                       np.linspace(-width / 2 + pixel_size/2, width / 2 - pixel_size/2, vector_resolution))

    im = image(sim, qty=qty, width=width, weight=weight, restrict_depth=restrict_depth,
               **kwargs)

    axes = kwargs.get('axes', None)
    if axes is None:
        axes = p.gca()

    if mode == 'quiver':
        if vector_scale is None:
            Q = axes.quiver(X, Y, vx, vy, color=vector_color, edgecolor=vector_edgecolor)
        else:
            if isinstance(vector_scale, str):
                vector_scale = _units.Unit(vector_scale)
            if _units.is_unit(vector_scale):
                vector_scale = vector_scale.in_units(sim['vel'].units)
            Q = axes.quiver(X, Y, vx, vy, scale=vector_scale, color=vector_color,
                         edgecolor=vector_edgecolor)
        if key:
            from . import quiverkey
            qk = quiverkey.PynbodyQuiverKey(Q, key_x, key_y,
                                      key_unit.in_units(sim['vel'].units, **sim.conversion_context()),
                                            "$" + key_unit.latex() + "$",
                                      color=key_color, labelcolor=key_color,
                                      boxedgecolor=key_edge_color, boxfacecolor=key_bg_color)
            qk.set_zorder(6)
            p.gca().add_artist(qk)
    elif mode == 'stream' :
        Q = axes.streamplot(X, Y, vx, vy, color=vector_color, density=stream_density, linewidth=stream_linewidth)

    axes.set_xlim(-width/2, width/2)
    axes.set_ylim(-width/2, width/2)


    return im


def volume(sim, qty='rho', width=None, resolution=200,
           color=(1.0,1.0,1.0),vmin=None,vmax=None,
           dynamic_range=4.0,log=True,
           create_figure=True):
    """Create a volume rendering of the given simulation using mayavi.

    **Keyword arguments:**

    *qty* (rho): The name of the array to interpolate

    *width* (None): The width of the cube to generate, centered on the origin

    *resolution* (200): The number of elements along each side of the cube

    *color* (white): The color of the volume rendering. The value of each voxel
       is used to set the opacity.

    *vmin* (None): The value for zero opacity (calculated using dynamic_range if None)

    *vmax* (None): The value for full opacity (calculated from the maximum
       value in the region if None)

    *dynamic_range*: The dynamic range to use if vmin and vmax are not specified

    *log* (True): log-scale the image before passing to mayavi

    *create_figure* (True): create a new mayavi figure before rendering
    """

    import mayavi
    from mayavi import mlab
    from tvtk.util.ctf import ColorTransferFunction, PiecewiseFunction



    if create_figure:
        fig = mlab.figure(size=(500,500),bgcolor=(0,0,0))

    grid_data = sph.to_3d_grid(sim,qty=qty,nx=resolution,
                               x2=None if width is None else width/2)



    if log:
        grid_data = np.log10(grid_data)
        if vmin is None:
            vmin = grid_data.max()-dynamic_range
        if vmax is None:
            vmax = grid_data.max()
    else:
        if vmin is None:
            vmin = np.min(grid_data)
        if vmax is None:
            vmax = np.max(grid_data)

    grid_data[grid_data<vmin]=vmin
    grid_data[grid_data>vmax]=vmax

    otf = PiecewiseFunction()
    otf.add_point(vmin,0.0)
    otf.add_point(vmax,1.0)

    sf = mayavi.tools.pipeline.scalar_field(grid_data)
    V = mlab.pipeline.volume(sf,color=color,vmin=vmin,vmax=vmax)




    V.trait_get('volume_mapper')['volume_mapper'].blend_mode = 'maximum_intensity'

    if color is None:
        ctf = ColorTransferFunction()
        ctf.add_rgb_point(vmin,107./255,124./255,132./255)
        ctf.add_rgb_point(vmin+(vmax-vmin)*0.8,200./255,178./255,164./255)
        ctf.add_rgb_point(vmin+(vmax-vmin)*0.9,1.0,210./255,149./255)
        ctf.add_rgb_point(vmax,1.0,222./255,141./255)

        V._volume_property.set_color(ctf)
        V._ctf = ctf
        V.update_ctf = True

    V._otf = otf
    V._volume_property.set_scalar_opacity(otf)


    return V


def _units_imply_projection(sim, qty, units):
    if isinstance(qty, str):
        qty = sim[qty]

    try:
        qty.units.ratio(units, **sim.conversion_context())
        # if this fails, perhaps we're requesting a projected image?
        return False
    except _units.UnitsException:
        # if the following fails, there's no interpretation this routine
        # can cope with. The error will be allowed to propagate.
        qty.units.ratio(
            units / (sim['x'].units), **sim.conversion_context())
        return True


def image(sim, qty='rho', width="10 kpc", resolution=None, units=None, log=True,
          vmin=None, vmax=None, weight=False, z_camera=None, clear=True, cmap=None,
          title=None, colorbar_label=None, qtytitle=None, show_cbar=True, axes=None,
          noplot=False, filename=None,
          return_image=False, return_array=False,
          fill_nan=True, fill_val=0.0, linthresh=None,
          restrict_depth = False,
          kernel_type='spline',
          **kwargs):
    """
    Make an image of the given simulation, using SPH or denoised-SPH interpolation.

    Parameters
    ----------
    sim : pynbody.snapshot.simsnap.SimSnap
        The simulation snapshot to plot. The image is generated in the plane z=0, or
        projected along the z axis.

    qty : str | pynbody.array.SimArray
        The name of the array to interpolate. Default is 'rho', which gives a density
        image. Alternatively, an array can be passed in.

    width : str or float, optional
        The overall width and height of the plot. If a float, it is assumed
        to be in units of sim['pos']. It can also be passed in as a string indicating
        the units, e.g. '10 kpc'. (Default is '10 kpc')

    resolution : int, optional
        The number of pixels wide and tall. (Default is determined by the
        :ref:`configuration file <configuration>`.)

    units : str or pynbody.units.Unit, optional
        The units of the output. Default is None, in which case the units of the input
        quantity are used. If the units correspond to integrating the quantity along
        a spatial dimension, the output is a projected image. For example, if the units
        are 'Msol kpc^-2', and the quantity is 'rho', the output is a projected image of
        the surface density.

    log : bool, optional
        If True, the image is log-scaled. (Default is True)

    vmin : float, optional
        The minimum value for the color scale. If None, the minimum value of the image.

    vmax : float, optional
        The maximum value for the color scale. If None, the maximum value of the image.

    weight : bool or str, optional
        If True, the requested quantity is volume-averaged down the line of sight. If a string, the
        requested quantity is averaged down the line of sight weighted by the named array
        (e.g. use 'rho' for density-weighted quantity).

    restrict_depth : bool, optional
        If True, restrict the depth of the image to the width of the image. (Default is False)

    z_camera : float, optional
        If set, a perspective image is rendered, as though the camera is a pinhole camera
        at (0,0,z_camera). The frustrum is defined by the width of the image in the plane z=0.

    clear : bool, optional
        Whether to clear the axes before plotting. (Default is True)

    cmap : matplotlib.colors.Colormap or str, optional
        Colormap to use. If None, the default colormap is used.

    title : str, optional
        Plot title.

    colorbar_label : str, optional
        Colorbar label. If not provided, one will be generated from the quantity name and units.

    show_cbar : bool, optional
        Whether to automatically plot the colorbar. (Default is True)

    axes : matplotlib.axes.Axes, optional
        Axes instance on which the image will be shown; if None, the current pyplot figure is
        used. (Default is False)

    noplot : bool, optional
        If True, the image is not displayed, only the image array is returned. This option therefore
        implies return_array = True.

    filename : str, optional
        If set, the image will be saved in a file.

    return_image : bool, optional
        If True, the image instance returned by imshow is returned. (Default is False)

    return_array : bool, optional
        If True, the numpy array of the image is returned. (Default is False)

    fill_nan : bool, optional
        If any of the image values are NaN, replace with fill_val. (Default is True)

    fill_val : float, optional
        The fill value to use when replacing NaNs. (Default is 0.0)

    linthresh : float, optional
        If the image has negative and positive values and a log scaling is requested, the part
        between -linthresh and linthresh is shown on a linear scale to avoid divergence at 0.

    kernel_type : str, optional
        SPH kernel to use for smoothing. Defaults to a cubic spline, but can also be set to 'wendlandC2'.
        (Default is 'spline')

    approximate_fast : bool, optional
        If True, speed up the image-making by rendering large kernels onto a lower-resolution image
        first, which is then interpolated to the final resolution. (Default is set in the config file).

    threaded : bool, optional
        If True, use threads to parallelise the rendering. (Default is set in the config file).

    qtytitle : str, optional
        Deprecated alias for colorbar_label.

    av_z : bool or str, optional
        Deprecated alias for weight.

    Returns
    -------

    matplotlib.image.AxesImage
        The image instance returned by imshow, if return_image is True.

    numpy.ndarray
        The image array, if return_array is True or noplot is True.

    """

    global config

    if resolution is None:
        resolution = config['image-default-resolution']

    if not noplot:
        import matplotlib.pylab as plt


    if not noplot:
        if axes:
            p = axes
        else:
            p = plt


    if qtytitle is not None:

        warnings.warn("qtytitle is deprecated; use colorbar_label instead", DeprecationWarning)
        colorbar_label = qtytitle

    if kwargs.get('av_z', None) is not None:
        weight = kwargs.pop('av_z')
        warnings.warn("av_z is deprecated; use weight instead", DeprecationWarning)

    if kwargs.get('ret_im', None) is not None:
        return_image = kwargs.pop('ret_im')
        warnings.warn("ret_im is deprecated; use return_image instead", DeprecationWarning)


    if colorbar_label is None and isinstance(qty, str):
        qtytitle = qty
    else:
        qtytitle = None

    if isinstance(units, str):
        units = _units.Unit(units)

    if isinstance(width, str) or isinstance(width, _units.UnitBase):
        if isinstance(width, str):
            width = _units.Unit(width)
        width = width.in_units(sim['pos'].units, **sim.conversion_context())

    width = float(width)

    if restrict_depth:
        kwargs['z_range'] = (-width/2, width/2)

    if kernel_type == 'wendlandC2':
        kernel = sph.WendlandC2Kernel()
    elif kernel_type == 'spline':
        kernel = sph.Kernel()
    else:
        raise ValueError('Invalid kernel_type specified. Options are "spline" (default) and "wendlandC2".')

    perspective = z_camera is not None
    if perspective and not weight:
        kernel = sph.Kernel2D(kernel)

    is_projected = False
    if units is not None:
        is_projected = _units_imply_projection(sim, qty, units)

    if is_projected:
        kernel = sph.Kernel2D(kernel)
        if qtytitle:
            qtytitle = f"$\\int\\,${qtytitle}$\\,\\mathrm{{d}}z$"

    if weight:
        if isinstance(kernel, sph.Kernel2D):
            raise _units.UnitsException(
                "Units already imply projected image; can't also average over line-of-sight!")
        if qtytitle:
            qtytitle = f"$\\langle${qtytitle}$\\rangle$"
        kernel = sph.Kernel2D(kernel)
        if units is not None:
            aunits = units * sim['z'].units
        else:
            aunits = None

        if isinstance(weight, str):
            if units is not None:
                aunits = units * sim[weight].units * sim['z'].units
            qty = sim[weight] * sim[qty]
        else:
            weight = np.ones_like(sim[qty])

        im = sph.render_image(sim, qty, width / 2, resolution, out_units=aunits, kernel=kernel,
                              z_camera=z_camera, **kwargs)
        im2 = sph.render_image(sim, weight, width / 2, resolution, kernel=kernel,
                               z_camera=z_camera, **kwargs)

        im = im / im2

    else:
        im = sph.render_image(sim, qty, width / 2, resolution, out_units=units,
                              kernel=kernel,  z_camera=z_camera, **kwargs)

    if fill_nan:
        im[np.isnan(im)] = fill_val

    if not noplot:

        # set the log or linear normalizations
        if log:
            try:
                im[np.where(im == 0)] = abs(im[np.where(abs(im != 0))]).min()
            except ValueError:
                raise ValueError("Failed to make a sensible logarithmic image. This probably means there are no particles in the view.")

            # check if there are negative values -- if so, use the symmetric
            # log normalization
            if (vmin is None and (im < 0).any() ) or ((vmin is not None) and vmin<0):

                # need to set the linear regime around zero -- set to by
                # default start at 1/1000 of the log range
                if linthresh is None:
                    linthresh = np.nanmax(abs(im)) / 1000.
                norm = matplotlib.colors.SymLogNorm(
                    linthresh, vmin=vmin, vmax=vmax)
            else:
                norm = matplotlib.colors.LogNorm(vmin=vmin, vmax=vmax)

        else:
            norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)

        #
        # do the actual plotting
        #
        if clear and not axes:
            p.clf()

        ims = p.imshow(im[::-1, :].view(np.ndarray), extent=(-width / 2, width / 2, -width / 2, width / 2),
                       cmap=cmap, norm = norm)

        u_st = sim['pos'].units.latex()
        if not axes:
            plt.xlabel("$x/%s$" % u_st)
            plt.ylabel("$y/%s$" % u_st)
        else:
            p.set_xlabel("$x/%s$" % u_st)
            p.set_ylabel("$y/%s$" % u_st)

        if units is None:
            units = im.units



        if units.latex() == "":
            units=""
        else:
            units = "$"+units.latex()+"$"

        if show_cbar:
            colorbar = plt.colorbar(ims)
            if colorbar_label is not None:
                colorbar.set_label(colorbar_label)
            elif qtytitle is not None:
                colorbar.set_label(qtytitle+"/"+units)


        if title is not None:
            if not axes:
                p.title(title)
            else:
                p.set_title(title)

        if filename is not None:
            p.savefig(filename)


    if return_image and return_array:
        return ims, im
    elif return_image:
        return ims
    elif return_array or noplot:
        return im



def image_radial_profile(im, bins=100):

    xsize, ysize = np.shape(im)
    x = np.arange(-xsize / 2, xsize / 2)
    y = np.arange(-ysize / 2, ysize / 2)
    xs, ys = np.meshgrid(x, y)
    rs = np.sqrt(xs ** 2 + ys ** 2)
    hist, bin_edges = np.histogram(rs, bins=bins)
    inds = np.digitize(rs.flatten(), bin_edges)
    ave_vals = np.zeros(bin_edges.size)
    max_vals = np.zeros(bin_edges.size)
    min_vals = np.zeros(bin_edges.size)
    for i in np.arange(bin_edges.size):
        try:
            min_vals[i] = np.min(10 ** (im.flatten()[np.where(inds == i)]))
        except ValueError:
            min_vals[i] = float('nan')
        ave_vals[i] = np.mean(10 ** (im.flatten()[np.where(inds == i)]))
        try:
            max_vals[i] = np.max(10 ** (im.flatten()[np.where(inds == i)]))
        except ValueError:
            max_vals[i] = float('nan')

    return ave_vals, min_vals, max_vals, bin_edges



import numpy as np

cimport cython
cimport numpy as np
cimport openmp
from cython.parallel cimport prange
from libc.math cimport atan, pow
from libc.stdlib cimport free, malloc

from .. import config


cdef extern from "geometry_selection.hpp" nogil:
    void wrapfn[T](T & x, T & y, T & z, const T & wrap, const T & wrap_by_two)
    void sphere_selection[T](T* position_array, char* output_array,
                              T x0, T y0, T z0, T max_radius,
                              T wrap,
                              Py_ssize_t num_particles, int num_threads)
    void cube_selection[T](T* position_array, char* output_array,
                            T x0, T y0, T z0, T x1, T y1, T z1,
                            T wrap,
                            Py_ssize_t num_particles, int num_threads)

ctypedef fused fused_float:
    np.float32_t
    np.float64_t

def selection(np.ndarray[fused_float, ndim=2] pos_ar, region, parameters, fused_float wrap):
    """OpenMP selection algorithm for spheres and cubes, optionally wrapped around the box.

    Parameters
    ----------
    pos_ar : array_like
        The positions of the particles, with shape (N,3). Must be C-contiguous.
        The datatype can either be float32 or float64, represented by fused_float.
    region : str
        The region to select particles from. Can be 'sphere' or 'cube'.
    parameters : array_like
        The parameters of the region. For a sphere, it is (x0, y0, z0, r_max), and for a cube,
        it is (x0, y0, z0, x1, y1, z1). Each parameter must be a float (will be cast to fused_float).
    wrap : fused_float
        The size of the box to wrap around. If <0, no wrapping is done.
    """

    cdef long i
    cdef long N=len(pos_ar)
    cdef fused_float cx,cy,cz,x,y,z,r2
    cdef fused_float r_max_2
    cdef np.ndarray[np.uint8_t, ndim=1] output = np.empty(len(pos_ar),dtype=np.uint8)

    cdef char* output_data = <char*> np.PyArray_DATA(output)

    if len(pos_ar) == 0:
        # The check for C-contiguous 3-col fails when len is zero, but there is nothing to do anyway, so just return
        return output

    if (np.PyArray_NDIM(pos_ar)!=2 or np.PyArray_DIMS(pos_ar)[1]!=3 or
            np.PyArray_STRIDES(pos_ar)[0] != sizeof(fused_float)*3 or
            np.PyArray_STRIDES(pos_ar)[1] != sizeof(fused_float)):
        raise ValueError("Input array must be C-contiguous and have 3 columns")

    cdef fused_float* pos_ar_data = <fused_float*> np.PyArray_DATA(pos_ar)

    cdef fused_float wrap_by_two = wrap/2
    cdef int num_threads = config['number_of_threads']

    if region == 'sphere':
        if len(parameters) != 4:
            raise ValueError("Sphere selection requires 4 parameters: (x0, y0, z0, r_max)")
        # (ugly syntax below is because starred parameters not allowed)
        sphere_selection(pos_ar_data, output_data,
                         <fused_float> parameters[0],
                         <fused_float> parameters[1],
                         <fused_float> parameters[2],
                         <fused_float> parameters[3], wrap,
                         len(pos_ar), num_threads)
    elif region == 'cube':
        if len(parameters) != 6:
            raise ValueError("Cube selection requires 6 parameters: (x0, y0, z0, x1, y1, z1)")
        cube_selection(pos_ar_data, output_data,
                       <fused_float> parameters[0],
                       <fused_float> parameters[1],
                       <fused_float> parameters[2],
                       <fused_float> parameters[3],
                       <fused_float> parameters[4],
                       <fused_float> parameters[5], wrap, len(pos_ar), num_threads)
    else:
        raise ValueError("Region must be either 'sphere' or 'cube'")

    return output



import importlib

from . import gas, generic, metals, profile, stars, util

importlib.reload(profile)
importlib.reload(generic)
importlib.reload(stars)
importlib.reload(gas)
importlib.reload(metals)
importlib.reload(util)

from .gas import rho_T, temp_profile
from .generic import fourier_map, gauss_kde, hist2d, qprof
from .metals import mdf, ofefeh
from .profile import density_profile, fourier_profile, rotation_curve
from .sph import faceon_image, image, sideon_image
from .stars import guo, satlf, sbprofile, schmidtlaw, sfh



import matplotlib.pylab as plt

import pynbody
import pynbody.plot.sph as sph

# load the snapshot and set to physical units
s = pynbody.load('testdata/gasoline_ahf/g15784.lr.01024.gz')
s.physical_units()

# load the halos
h = s.halos()

# center on the largest halo and align the disk
pynbody.analysis.angmom.faceon(h[1])

#create a simple slice of gas density
sph.image(h[1].g,qty="rho",units="g cm^-3",width=100,cmap="Greys")



"""Note that most simsnap tests are in more specific test files."""

import pathlib

import pytest

import pynbody


def test_load_empty_folder():

    pathlib.Path("testdata").joinpath('empty_folder_0001').mkdir(exist_ok=True)

    with pytest.raises(IOError) as excinfo:
        f = pynbody.load("testdata/empty_folder_0001")
    assert "Is a directory" in str(excinfo.value)



"""
Arrays for simulation data
==========================

The main class defined in this module is the `SimArray` class, which defines a shallow wrapper around ``numpy.ndarray``
for extra functionality like unit-tracking.

For most purposes, the differences between numpy.ndarray and
array.SimArray are not important. However, when units are specified
(by setting the ``units`` attribute), the behaviour is slightly
different. In particular,

* it becomes impossible to add or subtract arrays with incompatible dimensions

>>> SimArray([1,2], "Mpc") + SimArray([1,2], "Msol"))
ValueError

* addition or subtraction causes auto-unit conversion. For example

>>> SimArray([1,2], "Mpc") + SimArray([1,2], "kpc")
SimArray([1.001, 1.002], "Mpc")

* Note that in this context the left value takes precedence in
  specifying the return units, so that reversing the order of the
  operation here would return results in kpc.

* If only one of the arrays specifies a Unit, no checking occurs and
  the unit of the returned array is assumed to be the same as the one
  specified input unit.

* Powers to single integer or rational powers will maintain unit
  tracking.  Powers to float or other powers will not be able to do
  so.

>>> SimArray([1,2],"Msol Mpc**-3")**2
SimArray([1, 4], 'Msol**2 Mpc**-6')
>>> SimArray([1,2],"Msol Mpc**-3")**(1,3)
SimArray([ 1.,1.26], 'Msol**1/3 Mpc**-1')

Syntax above mirrors syntax in units module, where a length-two tuple
can represent a rational number, in this case one third.

>>> SimArray([1.,2], "Msol Mpc**-3")**0.333
SimArray([ 1.,1.26])  # Lost track of units



Getting the array in specified units
------------------------------------

Given an array, you can convert it in-place into units of your
own chosing:

>>> x = SimArray([1,2], "Msol")
>>> x.convert_units('kg')
>>> print(x)
SimArray([  1.99e+30,   3.98e+30], 'kg')

Or you can leave the original array alone and get a *copy* in
different units, correctly converted:

>>> x = SimArray([1,2], "Msol")
>>> print(x.in_units("kg"))
SimArray([  1.99e+30,   3.98e+30], 'kg')
>>> print(x)
SimArray([1,2], "Msol")

If the SimArray was created by a SimSnap (which is most likely), it
has a pointer into the SimSnap's properties so that the cosmological
context is automatically fetched. For example, comoving -> physical
conversions are correctly achieved:

>>> f = pynbody.load("fname")
>>> f['pos']
SimArray([[ 0.05419805, -0.0646539 , -0.15700017],
         [ 0.05169899, -0.06193341, -0.14475258],
         [ 0.05672406, -0.06384531, -0.15909944],
         ...,
         [ 0.0723075 , -0.07650762, -0.07657281],
         [ 0.07166634, -0.07453796, -0.08020873],
         [ 0.07165282, -0.07468577, -0.08020587]], '2.86e+04 kpc a')

>>> f['pos'].convert_units('kpc')
>>> f['pos']
SimArray([[ 1548.51403101, -1847.2525312 , -4485.71463308],
         [ 1477.1124212 , -1769.52421398, -4135.78377699],
         [ 1620.68592366, -1824.15000686, -4545.69387564],
         ...,
         [ 2065.9264273 , -2185.92982874, -2187.79225915],
         [ 2047.60759667, -2129.6537339 , -2291.6758134 ],
         [ 2047.2214441 , -2133.87693163, -2291.59406997]], 'kpc')


Specifying rules for ufunc's
-----------------------------

In general, it's not possible to infer what the output units from a given
ufunc should be. While numpy built-in ufuncs should be handled OK, other
ufuncs will need their output units defined (otherwise a numpy.ndarray
will be returned instead of our custom type.)

To do this, decorate a function with SimArray.ufunc_rule(ufunc). The function
you define should take the same number of parameters as the ufunc. These will
be the input parameters of the ufunc. You should return the correct units for
the output, or raise units.UnitsException (in the latter case, the return
array will be made into a numpy.ndarray.)

For example, here is the code for the correct addition/subtraction
handler:

.. code-block:: python

    @SimArray.ufunc_rule(np.add)
    @SimArray.ufunc_rule(np.subtract)
    def _consistent_units(a,b) :

        # This will be called whenever the standard numpy ufuncs np.add
        # or np.subtract are called with parameters a,b.

        # You should always be ready for the inputs to have no units.

        a_units = a.units if hasattr(a, 'units') else None
        b_units = b.units if hasattr(b, 'units') else None

        # Now do the logic. If we're adding incompatible units,
        # we want just to get a plain numpy array out. If we only
        # know the units of one of the arrays, we assume the output
        # is in those units.

        if a_units is not None and b_units is not None :
            if a_units==b_units :
                return a_units
            else :
                raise units.UnitsException("Incompatible units")

        elif a_units is not None :
            return a_units
        else :
            return b_units

"""

from __future__ import annotations

import fractions
import functools
import logging
import weakref
from typing import TYPE_CHECKING

import numpy as np

from .. import units

if TYPE_CHECKING:
    from .. import family, snapshot

_units = units

logger = logging.getLogger('pynbody.array')





class SimArray(np.ndarray):
    """A shallow wrapper around numpy.ndarray for extra functionality like unit-tracking."""

    _ufunc_registry = {}

    __slots__ = ['_units', '_sim', '_name', '_family']

    @property
    def ancestor(self):
        """Provides the basemost SimArray that an IndexedSimArray is based on."""
        return self

    @property
    def derived(self):
        """True if this array has been derived by pynbody; False otherwise.

        For more information on derived arrays, see :ref:`derived_arrays`."""
        if self.sim and self.name:
            return self.sim.is_derived_array(self.name, getattr(self, 'family', None))
        else:
            return False

    @derived.setter
    def derived(self, value):
        if value:
            raise ValueError("Can only unlink an array. Delete an array to force a rederivation if this is the intended effect.")
        if self.derived:
            self.sim.unlink_array(self.name)

    def __reduce__(self):
        T = np.ndarray.__reduce__(self)
        T = (
            T[0], T[1], (self.units, T[2][0], T[2][1], T[2][2], T[2][3], T[2][4]))
        return T

    def __setstate__(self, args):
        self._units = args[0]
        self.sim = None
        self._name = None
        np.ndarray.__setstate__(self, args[1:])

    def __init__(self, data, units = None, sim = None, **kwargs):
        """Initialise a SimArray with the specified units and simulation context.

        Other arguments are as for numpy.ndarray. If the data argument is a SimArray, the units and sim arguments are ignored."""

        # The actual initialisation is done by __new__ (it's actually unclear to me now why that should be the case,
        # but it's been like this for a long time and so changing it would need to be handled with care.)
        pass

    def __new__(cls, data, units=None, sim=None, **kwargs):
        new = np.array(data, **kwargs).view(cls)
        if hasattr(data, 'units') and hasattr(data, 'sim') and units is None and sim is None:
            units = data.units
            sim = data.sim

        if hasattr(data, 'family'):
            new.family = data.family

        if isinstance(units, str):
            units = _units.Unit(units)

        new._units = units

        # Always associate a SimArray with the top-level snapshot.
        # Otherwise we run into problems with how the reference should
        # behave: we don't want to lose the link to the simulation by
        # storing a weakref to a SubSnap that might be deconstructed,
        # but we also wouldn't want to store a strong ref to a SubSnap
        # since that would keep the entire simulation alive even if
        # deleted.
        #
        # So, set the sim attribute to the top-level snapshot and use
        # the normal weak-reference system.

        if sim is not None:
            new.sim = sim.ancestor
            # will generate a weakref automatically

        new._name = None

        return new

    def __array_finalize__(self, obj):
        if obj is None:
            return
        elif obj is not self and hasattr(obj, 'units'):
            self._units = obj.units
            self._sim = obj._sim
            self._name = obj._name
            if hasattr(obj, 'family'):
                self.family = obj.family
        else:
            self._units = None
            self._sim = lambda: None
            self._name = None

    def __array_function__(self, func, types, args, kwargs):
        ok_types = (SimArray, np.ndarray)
        if not all(issubclass(t, ok_types) for t in types):
            return NotImplemented

        args_processed = []
        for arg in args:
            if isinstance(arg, SimArray):
                args_processed.append(arg.view(np.ndarray))
            else:
                args_processed.append(arg)

        kwargs_processed= {}
        for k, v in kwargs.items():
            if isinstance(v, SimArray):
                kwargs_processed[k] = v.view(np.ndarray)
            else:
                kwargs_processed[k] = v

        result = func(*args_processed, **kwargs_processed)

        if func in SimArray._ufunc_registry:
            result = result.view(SimArray)
            if isinstance(result, SimArray): # may not be true if the result is a scalar
                sim = None
                for arg in args:
                    if isinstance(arg, SimArray):
                        sim = arg.sim
                        break
                result.units = SimArray._ufunc_registry[func](*args, **kwargs)
                result.sim = sim

        return result


    def __array_wrap__(self, array, context=None):
        if context is None:
            n_array = array.view(SimArray)
            return n_array

        try:
            ufunc = context[0]
            output_units = SimArray._ufunc_registry[ufunc](*context[1])
            n_array = array.view(SimArray)
            n_array.units = output_units
            n_array.sim = self.sim
            n_array._name = self._name
            return n_array
        except (KeyError, units.UnitsException):
            return array

    @staticmethod
    def ufunc_rule(for_ufunc):
        def x(fn):
            SimArray._ufunc_registry[for_ufunc] = fn
            return fn

        return x

    @property
    def units(self) -> units.UnitBase:
        """The units of the array, if known; otherwise :ref:`units.NoUnit`."""
        if hasattr(self.base, 'units'):
            return self.base.units
        else:
            if self._units is None:
                return _units.no_unit
            else:
                return self._units

    @units.setter
    def units(self, u):
        if not isinstance(u, units.UnitBase) and u is not None:
            u = units.Unit(u)

        if hasattr(self.base, 'units'):
            self.base.units = u
        else:
            if hasattr(u, "_no_unit"):
                self._units = None
            else:
                self._units = u

    @property
    def name(self) -> str | None:
        """The name of the array in the simulation snapshot, if known."""
        if hasattr(self.base, 'name'):
            return self.base.name
        return self._name

    @property
    def sim(self) -> snapshot.SimSnap | None:
        """The simulation snapshot that the array belongs to, if known."""
        if hasattr(self.base, 'sim'):
            base_sim = self.base.sim
        else:
            base_sim = self._sim()

        if self.family is not None and base_sim is not None:
            return base_sim[self.family]
        else:
            return base_sim


    @sim.setter
    def sim(self, s):
        if hasattr(self.base, 'sim'):
            self.base.sim = s
        else:
            if s is not None:
                self._sim = weakref.ref(s)
            else:
                self._sim = lambda: None

    @property
    def family(self) -> family.Family | None:
        """Returns the pynbody family that the array belongs to, if any.

        If ``family`` isn't None, an array ``a`` belongs to ``a.sim[family]`` rather than ``a.sim``.

        This doesn't necessarily mean that it is a family-level array, however, since it could be a slice into
        a simulation array.
        """
        try:
            return self._family
        except AttributeError:
            return None

    @family.setter
    def family(self, fam):
        self._family = fam

    def __mul__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            x = self.copy()
            x.units = x.units * rhs
            return x
        else:
            return np.ndarray.__mul__(self, rhs)

    def __div__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            x = self.copy()
            x.units = x.units / rhs
            return x
        else:
            return np.ndarray.__div__(self, rhs)

    def __truediv__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            x = self.copy()
            x.units = x.units / rhs
            return x
        else:
            return np.ndarray.__truediv__(self, rhs)

    def __imul__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            self.units *= rhs
        else:
            np.ndarray.__imul__(self, rhs)
            try:
                self.units *= rhs.units
            except AttributeError:
                pass
        return self

    def __idiv__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            self.units /= rhs
        else:
            np.ndarray.__idiv__(self, rhs)
            try:
                self.units /= rhs.units
            except AttributeError:
                pass
        return self

    def __itruediv__(self, rhs):
        if isinstance(rhs, _units.UnitBase):
            self.units /= rhs
        else:
            np.ndarray.__itruediv__(self, rhs)
            try:
                self.units /= rhs.units
            except AttributeError:
                pass
        return self

    def conversion_context(self):
        """Return a dictionary of contextual information that may be required for unit conversion.

        This is typically cosmological scalefactor and Hubble parameter."""
        if self.sim is not None:
            return self.sim.conversion_context()
        else:
            return {}

    def _generic_add(self, x, add_op=np.add):
        if hasattr(x, 'units') and not hasattr(self.units, "_no_unit") and not hasattr(x.units, "_no_unit"):
            # Check unit compatibility

            try:
                context = x.conversion_context()
            except AttributeError:
                context = {}

            # Our own contextual information overrides x's
            context.update(self.conversion_context())

            try:
                cr = x.units.ratio(self.units,
                                   **context)
            except units.UnitsException:
                raise ValueError("Incompatible physical dimensions {!r} and {!r}, context {!r}".format(
                    str(self.units), str(x.units), str(self.conversion_context())))

            if cr == 1.0:
                r = add_op(self, x)

            else:
                b = np.multiply(x, cr)
                if hasattr(b, 'units'):
                    b.units = None

                if not np.can_cast(b.dtype,self.dtype):
                    b = np.asarray(b, dtype=x.dtype)


                r = add_op(self, b)

            return r

        elif units.is_unit(x):
            x = x.in_units(self.units, **self.conversion_context())
            r = add_op(self, x)
            return r
        else:
            r = add_op(self, x)
            return r

    def __add__(self, x):
        if isinstance(x, _units.UnitBase):
            return x + self
        else:
            return self._generic_add(x)

    def __sub__(self, x):
        if isinstance(x, _units.UnitBase):
            return (-x + self).in_units(self.units)
        else:
            return self._generic_add(x, np.subtract)

    def __iadd__(self, x):
        self._generic_add(x, np.ndarray.__iadd__)
        return self

    def __isub__(self, x):
        self._generic_add(x, np.ndarray.__isub__)
        return self

    def __pow__(self, x):
        numerical_x = x

        if isinstance(x, tuple):
            x = fractions.Fraction(x[0], x[1])
            numerical_x = float(x)
        elif isinstance(x, fractions.Fraction):
            numerical_x = float(x)

        # The following magic circumvents our normal unit-assignment
        # code which couldn't cope with the numerical version of x
        # in the case of fractions. All this is necessary to make the
        # magic tuple->fraction conversions work seamlessly.
        r = np.asarray(np.power(self.view(np.ndarray), numerical_x)).view(SimArray)

        # Recent numpy versions can take 1-element arrays and return
        # scalars, in which case we now have a floating point number :(
        if type(r) is not SimArray:
            return r

        if self.units is not None and (
                isinstance(x, fractions.Fraction) or
                isinstance(x, int)):
            r.sim = self.sim
            r.units = self.units ** x
        else:
            r.units = None
            r.sim = self.sim

        return r

    def __repr__(self):
        x = np.ndarray.__repr__(self)
        if not hasattr(self.units, "_no_unit"):
            return x[:-1] + ", '" + str(self.units) + "')"
        else:
            return x

    def __setitem__(self, item, to):
        if hasattr(to, "in_units") and not hasattr(self.units, "_no_unit") and not hasattr(to.units, "_no_unit"):
            np.ndarray.__setitem__(self, item, to.in_units(self.units))
        else:
            np.ndarray.__setitem__(self, item, to)

    def __setslice__(self, a, b, to):
        self.__setitem__(slice(a, b), to)

    def abs(self, *args, **kwargs):
        x = np.abs(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def cumsum(self, axis=None, dtype=None, out=None):
        x = np.ndarray.cumsum(self, axis, dtype, out)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def prod(self, axis=None, dtype=None, out=None):
        x = np.ndarray.prod(self, axis, dtype, out)
        if hasattr(x, 'units') and axis is not None and self.units is not None:
            x.units = self.units ** self.shape[axis]
        if hasattr(x, 'units') and axis is None and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def sum(self, *args, **kwargs):
        x = np.ndarray.sum(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def mean(self, *args, **kwargs):
        x = np.ndarray.mean(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def mean_by_mass(self, *args, **kwargs):
        """Removed in pynbody 2.0. Use :func:`pynbody.snapshot.simsnap.SimSnap.mean_by_mass` instead."""
        raise RuntimeError("SimArray.mean_by_mass has been removed. Use SimSnap.mean_by_mass instead.")

    def max(self, *args, **kwargs):
        x = np.ndarray.max(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def min(self, *args, **kwargs):
        x = np.ndarray.min(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def ptp(self, *args, **kwargs):
        x = np.ndarray.ptp(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def std(self, *args, **kwargs):
        x = np.ndarray.std(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def var(self, *args, **kwargs):
        x = np.ndarray.var(self, *args, **kwargs)
        if hasattr(x, 'units') and self.units is not None:
            x.units = self.units ** 2
        if hasattr(x, 'sim') and self.sim is not None:
            x.sim = self.sim
        return x

    def set_units_like(self, new_unit):
        """Set the units of this array using a guess the ``sim``'s units for a given dimensionality.

        For example, if ``sim`` has units of ``Msol`` and ``kpc``, and ``new_unit`` is ``kg m^-3``, the
        units of this array will be set to ``Msol kpc^-3``.

        The underlying code uses dimensional analysis; of course, simulation codes are free to use inconsistent
        units if they like, so in general this routine cannot be guaranteed to infer the correct units. Human
        cross-checks are strongly advised.

        Note that this does not convert the array to the new units, it only sets the units attribute. To convert,
        use :func:`in_units`, :func:`convert_units` or :func:`in_original_units`.
        """

        if self.sim is not None:
            self.units = self.sim.infer_original_units(new_unit)
        else:
            raise RuntimeError("No link to SimSnap")

    def set_default_units(self, quiet=False):
        """Set the units for this array by guessing the ``sim``'s unit scheme and known dimensionality information.

        For example, if ``sim`` has units of ``Msol`` and ``kpc`` and this array is the ``rho`` array, the
        units of this array will be set to ``Msol kpc^-3``.

        The underlying code uses dimensional analysis; of course, simulation codes are free to use inconsistent
        units if they like, so in general this routine cannot be guaranteed to infer the correct units. Human
        cross-checks are strongly advised.

        Note that this does not convert the array to the new units, it only sets the units attribute. To convert,
        use :func:`in_units`, :func:`convert_units` or :func:`in_original_units`.
        """

        if self.sim is not None:
            try:
                self.units = self.sim._default_units_for(self.name)
            except (KeyError, units.UnitsException):
                if not quiet:
                    raise
        else:
            raise RuntimeError("No link to SimSnap")

    def in_original_units(self):
        """Return a copy of this array expressed in the file's internal unit scheme.

        For example, if ``sim`` has units of ``Msol`` and ``kpc`` and this array is the ``rho`` array, a copy of the
        array in units of ``Msol kpc^-3`` will be returned, even if the current units are something else like
        ``kg m^-3``.

        The underlying code uses dimensional analysis; of course, simulation codes are free to use inconsistent
        units if they like, so in general this routine cannot be guaranteed to infer the correct units. Human
        cross-checks are strongly advised.
        """

        return self.in_units(self.sim.infer_original_units(self.units))

    def in_units(self, new_unit, **context_overrides):
        """Return a copy of this array, expressed relative to an alternative unit ``new_unit``.

        Additional keyword arguments are interpreted as context overrides for the unit conversion. For example, if the
        array is a comoving distance and you want to convert it to a physical distance, you might call

        >>> x.in_units('kpc', a=0.1)

        to get the result assuming a scalefactor 0.1. If no context overrides are given, the context of the underlying
        ``sim`` is adopted.
        """

        context = self.conversion_context()
        context.update(context_overrides)

        if self.units is not None:
            r = self * self.units.ratio(new_unit,
                                        **context)
            r.units = new_unit
            return r
        else:
            raise ValueError("Units of array unknown")

    def convert_units(self, new_unit):
        """Convert units of this array in-place. If this is a sub-view, the entire base array will be converted."""

        if self.base is not None and hasattr(self.base, 'units'):
            self.base.convert_units(new_unit)
        else:
            ratio = self.units.ratio(new_unit,
                                     **(self.conversion_context()))
            logger.debug("Converting %s units from %s to %s; ratio = %.3e" %
                         (self.name, self.units, new_unit, ratio))
            self *= ratio
            self.units = new_unit

    def write(self, **kwargs):
        """
        Write this array to disk according to the standard method
        associated with its base file. This is equivalent to calling

        >>> sim.gas.write_array('array')

        in the case of writing out the array 'array' for the gas
        particle family.  See the description of
        :func:`pynbody.snapshot.SimSnap.write_array` for options.
        """

        if self.sim and self.name:
            self.sim.write_array(self.name, fam=self.family, **kwargs)
        else:
            raise RuntimeError("No link to SimSnap")


# Set up the correct comparison functions

def _unit_aware_comparison(ar, other, comparison_op=None):
    # guaranteed to be called with ar a SimArray instance
    if units.is_unit_like(other):
        if units.has_units(ar):
            # either other is a unit, or an array with a unit If
            # it's an array with a unit that matches our own, we
            # want to fall straight through to the comparison
            # operation. If it's an array with a unit that doesn't
            # match ours, OR it's a plain unit, we want to
            # convert first.
            if units.is_unit(other) or other.units != ar.units:
                other = other.in_units(ar.units)
        else:
            raise units.UnitsException("One side of a comparison has units and the other side does not")

    return comparison_op(ar, other)

for f in np.ndarray.__lt__, np.ndarray.__le__, np.ndarray.__eq__, \
        np.ndarray.__ne__, np.ndarray.__gt__, np.ndarray.__ge__:

    # N.B. cannot use functools.partial because it doesn't implement the descriptor
    # protocol
    @functools.wraps(f, assigned=("__name__", "__doc__"))
    def _wrapper_function(self, other, comparison_op=f):
        return _unit_aware_comparison(self, other, comparison_op=comparison_op)

    setattr(SimArray, f.__name__, _wrapper_function)


# Now add dirty bit setters to all the operations which are known
# to modify the numpy array

def _dirty_fn(w):
    def q(a, *y, **kw):
        if a.sim is not None and a.name is not None:
            a.sim._dirty(a.name)

        if kw != {}:
            return w(a, *y, **kw)
        else:
            return w(a, *y)

    q.__name__ = w.__name__
    return q

_dirty_fns = ['__setitem__', '__setslice__',
              '__irshift__',
              '__imod__',
              '__iand__',
              '__ifloordiv__',
              '__ilshift__',
              '__imul__',
              '__ior__',
              '__ixor__',
              '__isub__',
              '__invert__',
              '__iadd__',
              '__itruediv__',
              '__idiv__',
              '__ipow__']

for x in _dirty_fns:
    setattr(SimArray, x, _dirty_fn(getattr(SimArray, x)))

_u = SimArray.ufunc_rule


def _get_units_or_none(*a):
    if len(a) == 1:
        if hasattr(a[0], "units"):
            return a[0].units
        else:
            return None
    else:
        r = []
        for x in a:
            if hasattr(x, "units"):
                r.append(x.units)
            else:
                r.append(None)

        return r

#
# Now we have the rules for unit outputs after numpy built-in ufuncs
#
# Note if these raise UnitsException, a standard numpy array is returned
# from the ufunc to indicate the units can't be calculated. That means
# ufuncs can do 'non-physical' things, but then return ndarrays instead
# of SimArrays.


@_u(np.sqrt)
def _sqrt_units(a):
    if a.units is not None:
        return a.units ** (1, 2)
    else:
        return None


@_u(np.multiply)
def _mul_units(a, b):
    a_units, b_units = _get_units_or_none(a, b)
    if a_units is not None and b_units is not None:
        return a_units * b_units
    elif a_units is not None:
        return a_units
    else:
        return b_units


@_u(np.divide)
@_u(np.true_divide)
def _div_units(a, b):
    a_units, b_units = _get_units_or_none(a, b)
    if a_units is not None and b_units is not None:
        return a_units / b_units
    elif a_units is not None:
        return a_units
    else:
        return 1 / b_units


@_u(np.add)
@_u(np.subtract)
def _consistent_units(a, b):
    a_units, b_units = _get_units_or_none(a, b)
    if a_units is not None and b_units is not None:
        if a_units == b_units:
            return a_units
        else:
            raise units.UnitsException("Incompatible units")

    elif a_units is not None:
        return a_units
    else:
        return b_units


@_u(np.power)
def _pow_units(a, b):
    a_units = _get_units_or_none(a)
    if a_units is not None:
        if not isinstance(b, int) and not isinstance(b, units.Fraction):
            raise units.UnitsException("Can't track units")
        return a_units ** b
    else:
        return None


@_u(np.arctan)
@_u(np.arctan2)
@_u(np.arcsin)
@_u(np.arccos)
@_u(np.arcsinh)
@_u(np.arccosh)
@_u(np.arctanh)
@_u(np.sin)
@_u(np.tan)
@_u(np.cos)
@_u(np.sinh)
@_u(np.tanh)
@_u(np.cosh)
def _trig_units(*a):
    return 1


@_u(np.greater)
@_u(np.greater_equal)
@_u(np.less)
@_u(np.less_equal)
@_u(np.equal)
@_u(np.not_equal)
def _comparison_units(*a):
    return None

@_u(np.linalg.norm)
def _norm_units(a, *args, **kwargs):
    return a.units


class IndexedSimArray:
    """A view into a SimArray that allows for indexing and slicing.

    Unlike numpy arrays, IndexedSimArrays do not copy data when indexed. Instead, they provide a view into the original
    data. This is used by pynbody to provide a view into a subset of a simulation snapshot without copying the data,
    while making sure any changes to the data are reflected in the original snapshot.

    For most purposes, an IndexedSimArray should behave exactly like a SimArray. However, advanced users may want to
    understand more about performance implications and ways to optimize code. This can be found in the
    :ref:`performance` section of the documentation.
    """
    @property
    def derived(self):
        return self.base.derived

    @property
    def ancestor(self):
        return self.base.ancestor

    def __init__(self, array: SimArray, ptr: slice | np.ndarray):
        """Initialise an IndexedSimArray based on an underlying SimArray and a pointer into that array.

        The pointer can be a slice or an array of indexes."""
        self.base = array
        self._ptr = ptr

    def __array__(self, dtype=None):
        return np.asanyarray(self.base[self._ptr], dtype=dtype)

    def _reexpress_index(self, index):
        if isinstance(index, tuple) or (isinstance(index, list) and len(index) > 0 and hasattr(index[0], '__len__')):
            return (self._ptr[index[0]],) + tuple(index[1:])
        else:
            return self._ptr[index]

    def __getitem__(self, item):
        return self.base[self._reexpress_index(item)]

    def __setitem__(self, item, to):
        self.base[self._reexpress_index(item)] = to

    def __getslice__(self, a, b):
        return self.__getitem__(slice(a, b))

    def __setslice__(self, a, b, to):
        self.__setitem__(slice(a, b), to)

    def __repr__(self):
        return self.__array__().__repr__()  # Could be optimized

    def __str__(self):
        return self.__array__().__str__()  # Could be optimized

    def __len__(self):
        return len(self._ptr)

    def __reduce__(self):
        return SimArray(self).__reduce__()

    @property
    def shape(self):
        x = [len(self._ptr)]
        x += self.base.shape[1:]
        return tuple(x)

    @property
    def ndim(self):
        return self.base.ndim

    @property
    def units(self):
        return self.base.units

    @units.setter
    def units(self, u):
        self.base.units = u

    @property
    def sim(self):
        if self.base.sim is not None:
            return self.base.sim[self._ptr]
        else:
            return None

    @sim.setter
    def sim(self, s):
        self.base.sim = s

    @property
    def dtype(self):
        return self.base.dtype

    def conversion_context(self):
        return self.base.conversion_context()

    def set_units_like(self, new_unit):
        self.base.set_units_like(new_unit)

    def in_units(self, new_unit, **context_overrides):
        return IndexedSimArray(self.base.in_units(new_unit, **context_overrides), self._ptr)

    def convert_units(self, new_unit):
        self.base.convert_units(new_unit)

    def write(self, **kwargs):
        self.base.write(**kwargs)

    def prod(self):
        return np.array(self).prod()


# The IndexedSimArray class is now supplemented by wrapping all the
# standard numpy methods with a generated function which extracts an
# array realization of the subview before calling the underlying
# method.

def _wrap_fn(w):
    @functools.wraps(w)
    def q(s, *y,  **kw):
        return w(SimArray(s), *y, **kw)

    q.__name__ = w.__name__
    return q

# functions we definitely want to wrap, even though there's an existing
# implementation
_override = "__eq__", "__ne__", "__gt__", "__ge__", "__lt__", "__le__"

for x in set(np.ndarray.__dict__).union(SimArray.__dict__):
    _w = getattr(SimArray, x)
    if 'array' not in x and ((not hasattr(IndexedSimArray, x)) or x in _override) and hasattr(_w, '__call__') and x!="__buffer__":
        setattr(IndexedSimArray, x, _wrap_fn(_w))


def array_factory(dims: int | tuple, dtype: np.dtype, zeros: bool, shared: bool) -> SimArray:
    """Create an array of dimensions *dims* with the given numpy *dtype*.

    Parameters
    ----------

    dims : int or tuple
        The dimensions of the array.
    dtype : numpy.dtype
        The data type of the array.
    zeros : bool
        If True, the array is guaranteed to be zeroed.
    shared : bool
        If True, the array uses shared memory and can be efficiently shared across processes.

    """
    if not hasattr(dims, '__len__'):
        dims = (dims,)

    if shared:
        from . import shared
        ret_ar = shared.make_shared_array(dims, dtype, zeros)

    else:
        if zeros:
            ret_ar = np.zeros(dims, dtype=dtype).view(SimArray)
        else:
            ret_ar = np.empty(dims, dtype=dtype).view(SimArray)
    return ret_ar
